from typing import cast

import attrs
import numpy as np

__all__ = [
    'RigEvent',
    'CamEvent'
]


@attrs.define
class RigEvent:
    """container for riglog event type/value"""

    name: str
    """event name"""
    data: np.ndarray
    """2d array (T, V)"""

    def __attrs_post_init__(self):
        if len(self.time) != len(self.value):
            raise RuntimeError('')

    @property
    def time(self) -> np.ndarray:
        return self.data[:, 0]

    @property
    def value(self) -> np.ndarray:
        return self.data[:, 1]

    @property
    def start_time(self) -> float:
        return cast(float, self.time[0])

    @property
    def end_time(self) -> float:
        return cast(float, self.time[-1])

    @property
    def value_index(self) -> np.ndarray:
        """i.e., lap_value, which start accumulating from 1, then turn to 0-base"""
        if np.all(np.diff(self.value)) == 1:
            return (self.value - 1).astype(int)

        raise RuntimeError('incorrect operation')

    def with_pseudo_value(self, value: float) -> 'RigEvent':
        """pseudo value. i.e., for scatter"""
        t = self.time
        v = np.full_like(t, value, dtype=np.double)
        return attrs.evolve(self, data=np.column_stack([t, v]))


@attrs.frozen
class CamEvent(RigEvent):
    """container for riglog camera event type/value"""

    def __len__(self):
        """n_pulses"""
        return len(self.value)

    @property
    def n_pulses(self) -> int:
        """number of camera pulse"""
        return len(self)

    @property
    def fps(self) -> float:
        """approximate median fps in hz"""
        ret = np.median(1 / np.diff(self.time))
        return cast(float, ret)




"""
Facemap Wrapper
===========================

This module provide the result parser facemap
Including keypoints and pupil diameter data


Example of load the pupil area
--------------------------------

.. code-block:: python

    from neuralib.wrapper.facemap import *

    directory = ...  # directory with output *proc.npy
    fmap = FaceMapResult.load(directory, track_type='pupil')
    area = fmap.get_pupil_area()



Example of launch the GUI with the existed results
--------------------------------------------------------

.. code-block:: python

    from neuralib.wrapper.facemap import *

    directory = ...  # directory with output *proc.npy (pathlike)
    with_keypoints = False  # if have keypoint tracking
    env_name = ...  # conda env name with facemap package installed (str)
    FaceMapResult.launch_facemap_gui(directory, with_keypoints=with_keypoints, env_name=env_name)



Example of load the keypoint result
------------------------------------
TODO



"""
from .core import *



from __future__ import annotations

from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
from iblatlas.atlas import AllenAtlas
from iblatlas.regions import BrainRegions
from matplotlib.axes import Axes

from neuralib.util.io import IBL_CACHE_DIRECTORY
from neuralib.util.util_type import PathLike, ArrayLike

__all__ = [
    #
    'IBL_PLANE_TYPE',
    'IBL_MAPPING_TYPE',
    'HEMISPHERE_TYPE',
    'IBL_BG_TYPE',
    #
    'IBLAtlasPlotWrapper'
]

IBL_PLANE_TYPE = Literal['coronal', 'sagittal', 'horizontal', 'top']
IBL_MAPPING_TYPE = Literal['Allen', 'Beryl', 'Cosmos', 'Swanson']  # merge areas
HEMISPHERE_TYPE = Literal['left', 'right', 'both']
IBL_BG_TYPE = Literal['image', 'boundary']


class IBLAtlasPlotWrapper(AllenAtlas):
    """Wrapper for iblatlas plotting usage

    .. seealso:: `<https://int-brain-lab.github.io/iblenv/atlas_examples.html>`_
    """
    regions: BrainRegions

    def __init__(self, res_um: int = 10,
                 scaling: tuple[int, int, int] = (1, 1, 1),
                 mock: bool = False,
                 source_path: PathLike | None = None,
                 alpha: float = 1.0):
        """

        :param res_um: The Atlas resolution in micrometres; one of 10, 25 or 50um
        :param scaling: Scale factor along ml, ap, dv for squeeze and stretch (default: [1, 1, 1])
        :param mock: For testing purposes, return atlas object with image comprising zeros
        :param source_path: The location of the image volume. By default, use ``IBL_CACHE_DIRECTORY``
        :param alpha:
        """

        if source_path is None:
            source_path = IBL_CACHE_DIRECTORY

        super().__init__(res_um, scaling, mock, source_path)

        self._alpha = alpha

    def get_acronym_list(self, mapping: IBL_MAPPING_TYPE) -> list[str]:
        """get acronym list"""
        # noinspection PyUnresolvedReferences
        return list(self.regions.acronym[self.regions.mappings[mapping]])

    def plot_scalar_on_slice(
            self,
            regions: ArrayLike, *,
            values: ArrayLike | None = None,
            coord: int = -570,
            plane: IBL_PLANE_TYPE = 'coronal',
            mapping: IBL_MAPPING_TYPE = 'Allen',
            hemisphere: HEMISPHERE_TYPE = 'left',
            background: IBL_BG_TYPE = 'image',
            cmap: str = 'viridis',
            clevels: ArrayLike | None = None,
            show_cbar: bool = False,
            empty_color: str = 'silver',
            ax: Axes | None = None,
            vector: bool = False,
            slice_files: np.ndarray | None = None,
            auto_merge: bool = True,
            verbose: bool = True,
            **kwargs
    ) -> tuple[plt.Figure, Axes] | tuple[plt.Figure, Axes, plt.colorbar]:
        """
        Plot the slice view.

        See detail in ``iblatlas.plots.plot_scalar_on_slice``

        .. seealso:: `<https://int-brain-lab.github.io/iblenv/notebooks_external/atlas_plotting_scalar_on_slice.html>`_

        :param regions:
        :param values:
        :param coord:
        :param plane:
        :param mapping:
        :param hemisphere:
        :param background:
        :param cmap:
        :param clevels:
        :param show_cbar:
        :param empty_color:
        :param ax:
        :param vector:
        :param slice_files:
        :param auto_merge: merge for the certain mapping
        :param verbose:
        :param kwargs:
        :return:
        """

        from iblatlas.plots import plot_scalar_on_slice

        if not isinstance(regions, np.ndarray):
            regions = np.array(regions)

        if values is None:
            values = np.arange(regions.size)

        if auto_merge:
            regions = np.unique(self.regions.acronym2acronym(regions, mapping=mapping))
            values = np.arange(regions.size)

        if verbose:
            print(f'Plot region: {list(regions)}')

        # noinspection PyTypeChecker
        return plot_scalar_on_slice(
            regions, values, coord, plane, mapping,
            hemisphere, background, cmap, clevels, show_cbar, empty_color,
            self, ax, vector, slice_files, **kwargs
        )

    def plot_points_on_slice(
            self,
            xyz: np.ndarray,
            values: ArrayLike | None = None,
            coord: int = -570,
            plane: IBL_PLANE_TYPE = 'coronal',
            mapping: IBL_MAPPING_TYPE = 'Allen',
            background: IBL_BG_TYPE = 'boundary',
            cmap: str = 'Reds',
            clevels: ArrayLike | None = None,
            show_cbar: bool = True,
            aggr: Literal['sum', 'count', 'mean', 'std', 'median', 'min', 'max'] = 'count',
            fwhm: int = 0,
            ax: Axes | None = None,
    ) -> tuple[plt.Figure, Axes] | tuple[plt.Figure, Axes, plt.colorbar]:
        """
        plot_points_on_slice

        :param xyz:
        :param values:
        :param coord:
        :param plane:
        :param mapping:
        :param background:
        :param cmap:
        :param clevels:
        :param show_cbar:
        :param aggr:
        :param fwhm:
        :param ax:
        :return:
        """
        from iblatlas.plots import plot_points_on_slice

        return plot_points_on_slice(
            xyz, values, coord, plane, mapping,
            background, cmap, clevels, show_cbar, aggr, fwhm,
            self, ax
        )

    def _plot_slice(self, im, extent, ax=None, cmap=None, volume=None, **kwargs):
        """Overwrite by pass other kwargs"""
        if ax is None:
            ax = plt.gca()
            ax.axis('equal')

        if cmap is None:
            cmap = plt.get_cmap('bone')

        if volume == 'boundary':
            imb = np.zeros((*im.shape[:2], 4), dtype=np.uint8)
            imb[im == 1] = np.array([0, 0, 0, 255])
            im = imb

        ax.imshow(im, extent=extent, cmap=cmap, alpha=self._alpha)



import random
from pathlib import Path

from neuralib.persistence import persistence


@persistence.persistence_class
class Example:
    a: int = persistence.field(validator=True, filename=True)
    b: int = persistence.autoinc_field()
    c: int

    def compute(self) -> 'Example':
        self.c = random.randint(0, self.a)
        return self

    def with_b(self, b: int) -> 'Example':
        self.b = b
        return self


handler = persistence.PickleHandler(Example, Path('../../rscvp2/test'))
print(persistence.as_dict(handler.save_persistence(Example(10).compute())))
print(persistence.as_dict(handler.load_persistence(Example(10).with_b(1))))



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import NamedTuple, Optional, Literal, Union, TypedDict

import numpy as np
import polars as pl
from scipy.ndimage import gaussian_filter1d

from neuralib.plot import plot_figure
from neuralib.util.util_type import PathLike
from neuralib.wrapper.deeplabcut.util import remove_jumps, interpolate_gaps, compute_velocity

__all__ = [
    'VideoConfig',
    'DlcResult',
    'DlcTrackPart'
]


# ============ #
# Video Config #
# ============ #

class VideoConfig(TypedDict):
    xy_pixel: tuple[int, int]
    width: int
    height: int
    length_unit: Literal['cm', 'mm']
    fps: float
    """in hz"""


def get_default_config() -> VideoConfig:
    return VideoConfig(
        xy_pixel=(1024, 1024),
        width=300,
        height=300,
        length_unit='mm',
        fps=30
    )


def read_default_config() -> VideoConfig:
    """TODO from video"""
    pass


def get_pixel2length_factor(config: VideoConfig) -> float:
    """"rough calculation"""
    wp = config['width'] / config['xy_pixel'][0]
    hp = config['height'] / config['xy_pixel'][1]
    return (wp + hp) / 2


# ========== #
# DLC Result #
# ========== #

@dataclasses.dataclass
class DlcResult:
    prefix: str
    model_name: str
    name: str
    code: str
    with_filter: bool

    config: VideoConfig

    data: pl.DataFrame
    """
    ┌────────────┬────────────┬─────────────────┬────────────┬────────────┬────────────────┬───────┐
    │ blue_x     ┆ blue_y     ┆ blue_likelihood ┆ red_x      ┆ red_y      ┆ red_likelihood ┆ *valid│
    │ ---        ┆ ---        ┆ ---             ┆ ---        ┆ ---        ┆ ---            ┆ ---   │
    │ f64        ┆ f64        ┆ f64             ┆ f64        ┆ f64        ┆ f64            ┆ bool  │
    ╞════════════╪════════════╪═════════════════╪════════════╪════════════╪════════════════╪═══════╡
    │ *NaN       ┆ NaN        ┆ 0.8             ┆ NaN        ┆ NaN        ┆ 1.0            ┆ false │
    │ 422.874268 ┆ 266.782867 ┆ 1.0             ┆ 422.117218 ┆ 272.525543 ┆ 1.0            ┆ true  │
    │ …          ┆ …          ┆ …               ┆ …          ┆ …          ┆ …              ┆ …     │
    │ 350.320679 ┆ 296.086517 ┆ 1.0             ┆ 357.244385 ┆ 291.834351 ┆ 1.0            ┆ true  │
    └────────────┴────────────┴─────────────────┴────────────┴────────────┴────────────────┴───────┘
    """

    parts: tuple[str, ...]

    valid_mask: Optional[np.ndarray] = dataclasses.field(init=False, default=None)
    invalid_frame_index: set[int] = dataclasses.field(default_factory=set)
    _col_order: tuple[str] = dataclasses.field(init=False, default=('x', 'y', 'likelihood'))

    def __post_init__(self):
        mask = np.ones(self.data.shape[0])
        invalid = np.array(list(self.invalid_frame_index)).astype(int)
        mask[invalid] = 0
        self.valid_mask = np.array(mask, dtype=bool)

    @classmethod
    def load(cls, filepath: PathLike,
             config: Optional[VideoConfig] = None) -> 'DlcResult':
        file = Path(filepath)
        p, m, n, c = file.stem.split('_')
        with_filter = True if c.endswith('filter') else False

        with file.open() as f:
            f.readline()
            parts = f.readline().strip().split(',')[1::3]

        new_col = ['']
        for p in parts:
            for it in cls._col_order:
                new_col.append(f'{p}_{it}')
        #
        data = pl.read_csv(filepath, skip_rows=3, has_header=False, new_columns=new_col)[:, 1:]

        #
        if config is None:
            config = get_default_config()

        #
        factor = get_pixel2length_factor(config)
        for c in ['x', 'y']:
            for p in parts:
                data = data.with_columns(pl.col(f'{p}_{c}') * factor)

        return DlcResult(p, m, n, c, with_filter, config, data, tuple(parts))

    def with_global_likelihood_filtered(self,
                                        threshold: float = 0.9,
                                        interp_nan: bool = True,
                                        interp_gap: float = 0.5) -> 'DlcResult':
        """For all parts
        replace xy as nan if any parts with low likelihood
        ** Note if even interp_nan, the non-interpolate nan need to be further processed in the caller site.
        """
        data = self.data.with_columns(
            pl.concat_list([pl.col(f'{p}_likelihood') >= threshold for p in self.parts]).list.all().alias('valid')
        )

        invalid = data.select(~pl.col('valid')).to_numpy().ravel()
        self.invalid_frame_index.update(np.nonzero(invalid)[0])

        data = data.with_columns(
            *([
                  pl.when(pl.col('valid')).then(pl.col(f'{p}_x')).otherwise(np.nan)
                  for p in self.parts
              ] + [
                  pl.when(pl.col('valid')).then(pl.col(f'{p}_y')).otherwise(np.nan)
                  for p in self.parts
              ])
        )

        if interp_nan:
            for p in self.parts:
                xy = data.select([pl.col(f'{p}_x', f'{p}_y')]).to_numpy()
                _xy = interpolate_gaps(self.time, xy, interp_gap)
                for i, c in enumerate(['x', 'y']):
                    data = data.with_columns(pl.Series(_xy[:, i]).alias(f'{p}_{c}'))

        return dataclasses.replace(self, data=data)

    def average_across(self, parts: list[str]) -> DlcTrackPart:
        if any([p not in self.parts for p in parts]):
            raise ValueError('')

        data = self.data.select(
            avg_x=pl.concat_list([f'{p}_x' for p in parts]).list.mean(),
            avg_y=pl.concat_list([f'{p}_y' for p in parts]).list.mean(),
            likelihood=pl.concat_list([f'{p}_likelihood' for p in parts]).list.mean()
        ).to_numpy()

        return DlcTrackPart(
            self,
            parts,
            data[:, :2],
            data[:, 2],
            self.config['fps']
        )

    def __getitem__(self, part: str) -> DlcTrackPart:
        if part not in self.parts:
            raise KeyError(f'unknown {part}, should be in {self.parts}')

        ret = self.data.select([f'{part}_{i}' for i in self._col_order]).to_numpy()

        return DlcTrackPart(
            self,
            part,
            ret[:, :2],
            ret[:, 2],
            self.fps
        )

    @property
    def n_frames(self) -> int:
        return self.data.shape[0]

    @property
    def time(self) -> np.ndarray:
        """Assume stable frame rate"""
        return np.linspace(0, self.total_time, self.n_frames)

    @property
    def total_time(self) -> float:
        return self.n_frames * self.fps

    @property
    def fps(self) -> float:
        return self.config['fps']

    @fps.setter
    def fps(self, value: float):
        if value != self.fps:
            print(f'reset video fps {self.fps} -> {value}')
        self.config['fps'] = value


# ========== #
# Track Part #
# ========== #

class DlcTrackPart(NamedTuple):
    source: 'DlcResult'

    parts: Union[str, list[str]]  # 1|P
    xy: np.ndarray  # (N, 2)
    likelihood: np.ndarray
    fps: float

    #
    processed: bool = False
    likelihood_removal_index: list[int] = []
    jump_removal_index: list[list[int]] = []

    @property
    def n_frames(self) -> int:
        return self.xy.shape[0]

    @property
    def time(self) -> np.ndarray:
        return np.linspace(0, self.total_time, self.n_frames)

    @property
    def total_time(self) -> float:
        return self.n_frames * self.fps

    @property
    def speed(self) -> np.ndarray:
        return np.abs(compute_velocity(self.xy, self.fps))

    @property
    def direction(self) -> np.ndarray:
        return np.angle(compute_velocity(self.xy, self.fps), deg=True)

    def with_likelihood_filtered(self, threshold: float = 0.95) -> 'DlcTrackPart':
        xy = self.xy.copy()
        invalid = np.nonzero(self.likelihood < threshold)[0]
        xy[invalid] = np.nan
        return self._replace(xy=xy, likelihood_removal_index=invalid.tolist())

    def with_jump_removal(self, jump_size: int = 20, duration: float = 0.1) -> 'DlcTrackPart':
        ret, jump = remove_jumps(self.time, self.xy, jump_size, duration)
        return self._replace(xy=ret, jump_removal_index=jump)

    def interpolate_gaps(self, interpolate_gap: float) -> 'DlcTrackPart':
        xy = interpolate_gaps(self.time, self.xy, interpolate_gap)
        return self._replace(xy=xy)

    def with_proc(self, jump_size: int = 20,
                  duration: float = 0.1,
                  interpolate_gap: float = 10) -> 'DlcTrackPart':
        if self.processed:
            return self
        return (self.with_likelihood_filtered()
                .with_jump_removal(jump_size, duration)
                .interpolate_gaps(interpolate_gap)
                ._replace(processed=True))

    def _kernel_filter(self,
                       xy: np.ndarray,
                       gaussian_kernel: Optional[float] = None,
                       run_kalman_filter: bool = False,
                       **kalman_kw) -> np.ndarray:

        if kalman_kw is None:
            kalman_kw = dict(P=200, R=0.5, Q=2000)  # TODO opt and not test yet

        if gaussian_kernel is not None:
            xy = gaussian_filter1d(xy, gaussian_kernel, axis=0)

        if run_kalman_filter:
            from rscvp.util.util_signal import run_kalman_2d
            speed = np.vstack([self.speed, self.speed]).T
            xs = run_kalman_2d(xy, speed, dt=1 / self.fps, **kalman_kw)[0]
            xy = xs[:, :2]

        return xy

    def plot_trajectory_occ(self, **kwargs):
        import seaborn as sns
        if not self.processed:
            xy = self.with_proc().xy
        else:
            xy = self.xy

        xy = self._kernel_filter(xy, **kwargs)
        with plot_figure(None, 1, 2) as ax:
            ax[0].plot(xy[:, 0], xy[:, 1])
            sns.histplot(x=xy[:, 0], y=xy[:, 1], cmap='twilight_shifted',
                         bins=100, ax=ax[1], cbar=True, cbar_kws=dict(shrink=0.25))



import abc
from pathlib import Path
from typing import TypedDict, Literal, Union, Optional, final, Any

import attrs
import cv2
import napari
import numpy as np
from typing_extensions import Self

from neuralib.argp import argument, as_argument
from neuralib.segmentation.base import AbstractSegmentationOption
from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint

__all__ = ['CPOSE_MODEL',
           'AbstractSegmentationOption',
           'AbstractCellPoseOption',
           'CellPoseEvalResult']


class ChannelDict(TypedDict):
    none: int
    gray: int
    red: int
    green: int
    blue: int


DEFAULT_CHANNEL_DICT: ChannelDict = {
    'none': -1,
    'gray': 0,
    'red': 1,
    'green': 2,
    'blue': 3
}

CPOSE_MODEL = Literal['cyto', 'cyto2', 'cyto3']


class AbstractCellPoseOption(AbstractSegmentationOption, metaclass=abc.ABCMeta):
    DESCRIPTION = 'ABC for GUI Cellpose'

    model: CPOSE_MODEL = as_argument(AbstractSegmentationOption.model).with_options(default='cyto3')

    chan_seg: str = argument(
        '-C', '--chan',
        help=f'channel for segmentation default:{DEFAULT_CHANNEL_DICT}'
    )

    chan_nuclear: str = argument(
        '-N', '--chan-nuclear',
        default=DEFAULT_CHANNEL_DICT['blue'],
        help='nuclear channel'
    )

    diameter: int = argument(
        '-D', '--diameter',
        default=7,
        help='diameter for each neuron (number of each pixel)'
    )

    cellpose_gui: bool = argument(
        '--cpose-gui',
        help='launch_cellpose_gui for the analyzed result'
    )

    @property
    def seg_result(self) -> Path:
        if not self.file.is_file():
            raise ValueError(f'{self.file} must be a file')
        return self.file.with_name(self.file.stem + '_seg').with_suffix('.npy')

    @abc.abstractmethod
    def eval(self) -> Optional['CellPoseEvalResult']:
        pass

    # noinspection PyTypeChecker
    def launch_napari(self):
        res = CellPoseEvalResult.load(self.seg_result)

        viewer = napari.Viewer()
        viewer.add_image(res.image, name='raw', colormap='cyan')
        viewer.add_image(res.masks, name='mask', opacity=0.5, colormap='red')
        viewer.add_image(res.outlines, name='outline', opacity=0.5)

        napari.run()

    def launch_cellpose_gui(self):
        """FIXME AttributeError: 'MainW' object has no attribute 'load_3D'. Cellpose version 3.0.5
        """
        from cellpose.gui.gui import run
        run(str(self.seg_result))


# ========== #
# EvalResult #
# ========== #
# TODO check doc and type in up-to date version
# TODO 3D image need to be tested

class NormParams(TypedDict):
    lowhigh: Optional[Any]
    percentile: list[float, float]
    normalize: bool
    norm3D: bool
    sharpen_radius: float
    smooth_radius: float
    tile_norm_blocksize: float
    tile_norm_smooth3D: float
    invert: bool


@final
@attrs.define
class CellPoseEvalResult:
    # inputs
    filename: str
    """image file name"""
    image: Union[np.ndarray, list[np.ndarray]]
    diameter: float
    """neuronal diameter"""
    chan_choose: list[int]
    """[chan_neg, chan_nuclear]"""

    # result
    masks: list[np.ndarray] = attrs.Factory(list)
    """each pixel in the image is assigned to an ROI (H, W)
    list of 2D arrays, labelled image, where 0=no masks; 1,2,...=mask labels
    """

    flows: list[np.ndarray] = attrs.Factory(list)
    """
    flows[0] is XY flow in RGB, 
    flows[1] is the cell probability in range 0-255 instead of 0.0 to 1.0, 
    flows[2] is Z flow in range 0-255 (if it exists, otherwise zeros),
    flows[3] is [dY, dX, cellprob] (or [dZ, dY, dX, cellprob] for 3D),
    flows[4] is pixel destinations (for internal use)
    """

    styles: list[np.ndarray] = attrs.Factory(list)
    """list of 1D arrays of length 256, 
    style vector summarizing each image, also used to estimate size of objects in image"""

    # ==================== #
    # Optional & Overwrite #
    # ==================== #

    # GUI dependent
    colors: Optional[np.ndarray] = attrs.field(default=None, kw_only=True)
    """colors for ROIs (N, 3)"""

    manual_changes: Optional[list[Any]] = attrs.field(default=None, kw_only=True)

    # CLI dependent
    est_diam: Optional[float] = attrs.field(default=None, kw_only=True)
    """estimated diameter (if run on command line)"""

    #
    model_path: int = attrs.field(default=0, kw_only=True)
    flow_threshold: Optional[float] = attrs.field(default=None, kw_only=True)
    cellprob_threshold: float = attrs.field(default=0, kw_only=True)
    normalize_params: Optional[NormParams] = attrs.field(default=None, kw_only=True)

    # restore
    img_restore: Optional[list[np.ndarray]] = attrs.field(default=None, kw_only=True)
    restore: Optional[str] = attrs.field(default=None, kw_only=True)
    ratio: float = attrs.field(default=1.0, kw_only=True)

    # Overwrite while calling save
    outlines: np.ndarray = attrs.field(default=np.array([]), kw_only=True)
    """outlines of ROIs (H, W)"""
    ismanual: np.ndarray = attrs.field(default=np.array([]), kw_only=True)
    """whether or not mask k was manually drawn or computed by the cellpose algorithm. bool array (N, )"""

    @classmethod
    def load(cls, seg_file: PathLike) -> Self:
        if not isinstance(seg_file, Path):
            seg_file = Path(seg_file)

        res = np.load(seg_file, allow_pickle=True).item()

        image_file = Path(res['filename'])
        if not image_file.exists():
            fprint(f'No image data found {image_file}', vtype='warning')
            image = None
        else:
            image = cv2.imread(res['filename'])

        return cls(**res, image=image)

    def save_seg_file(self) -> None:
        from cellpose.io import masks_flows_to_seg

        if not isinstance(self.image, list):
            self.image = [self.image]

        masks_flows_to_seg(self.image,
                           self.masks,
                           self.flows,
                           self.filename,
                           diams=self.diameter,
                           imgs_restore=self.img_restore,
                           restore_type=self.restore,
                           ratio=self.ratio)



from __future__ import annotations

import io
from io import BytesIO
from pathlib import Path
from pprint import pprint
from typing import ClassVar, TypedDict, Literal

import allensdk.core.structure_tree
import numpy as np
import pandas as pd
import polars as pl
from allensdk.core.reference_space_cache import ReferenceSpaceCache

from neuralib.atlas.util import PLANE_TYPE, ALLEN_SOURCE_TYPE
from neuralib.atlas.view import AbstractSliceView
from neuralib.util.io import CCF_CACHE_DIRECTORY
from neuralib.util.tqdm import download_with_tqdm
from neuralib.util.util_type import PathLike, DataFrame
from neuralib.util.util_verbose import fprint
from neuralib.util.utils import uglob

__all__ = ['AllenReferenceWrapper',
           'create_allen_structure_dict']


class StructureTreeDict(TypedDict):
    """allen structure tree dict"""
    acronym: str
    graph_id: int
    graph_order: int
    id: int
    name: str
    structure_id_path: list[int]
    structure_set_ids: list[int]
    rgb_triplet: list[int]


class AllenReferenceWrapper:
    """Class for load the data in allen"""

    REFERENCE_SPACE_KEY: ClassVar[str] = 'ccf_2017'
    STRUCTURE_GRAPH_ID: ClassVar[int] = 1

    def __init__(self,
                 resolution: int = 10,
                 output: PathLike | None = None):
        """
        :param resolution: resolution in um
        :param output: output directory for downloading cache. By default, save in ``CCF_CACHE_DIRECTORY``
        """

        if output is None:
            output = CCF_CACHE_DIRECTORY

        if not output.exists():
            output.mkdir(parents=True, exist_ok=True)

        self.source_root = Path(output)
        self.reference = ReferenceSpaceCache(resolution,
                                             self.REFERENCE_SPACE_KEY,
                                             manifest=self.source_root / 'manifest.json')

    # ============== #
    # Structure Tree #
    # ============== #

    @classmethod
    def load_structure_tree(cls, version: Literal['2017', 'old'] = '2017') -> pl.DataFrame:
        """
        Load structure tree dataframe

        :param version: {'2017', 'old'}
        :return: structure tree dataframe

        ::

            ┌───────────┬──────────┬───────────────────────────────┬─────────┬──────────┬─────────────┬───────────────┬────────┬─────────────────────┬───────┬──────────┬─────────────┬───────────────────┬───────────────────┬─────────────────────────┬──────────────────────────────┬────────┬───────────┬──────────────────────┬──────────────┬───────────────────────────────┐
            │ id        ┆ atlas_id ┆ name                          ┆ acronym ┆ st_level ┆ ontology_id ┆ hemisphere_id ┆ weight ┆ parent_structure_id ┆ depth ┆ graph_id ┆ graph_order ┆ structure_id_path ┆ color_hex_triplet ┆ neuro_name_structure_id ┆ neuro_name_structure_id_path ┆ failed ┆ sphinx_id ┆ structure_name_facet ┆ failed_facet ┆ safe_name                     │
            │ ---       ┆ ---      ┆ ---                           ┆ ---     ┆ ---      ┆ ---         ┆ ---           ┆ ---    ┆ ---                 ┆ ---   ┆ ---      ┆ ---         ┆ ---               ┆ ---               ┆ ---                     ┆ ---                          ┆ ---    ┆ ---       ┆ ---                  ┆ ---          ┆ ---                           │
            │ i64       ┆ f64      ┆ str                           ┆ str     ┆ str      ┆ i64         ┆ i64           ┆ i64    ┆ f64                 ┆ i64   ┆ i64      ┆ i64         ┆ list[i64]         ┆ str               ┆ str                     ┆ str                          ┆ str    ┆ i64       ┆ i64                  ┆ i64          ┆ str                           │
            ╞═══════════╪══════════╪═══════════════════════════════╪═════════╪══════════╪═════════════╪═══════════════╪════════╪═════════════════════╪═══════╪══════════╪═════════════╪═══════════════════╪═══════════════════╪═════════════════════════╪══════════════════════════════╪════════╪═══════════╪══════════════════════╪══════════════╪═══════════════════════════════╡
            │ 997       ┆ -1.0     ┆ root                          ┆ root    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ -1.0                ┆ 0     ┆ 1        ┆ 0           ┆ [997]             ┆ FFFFFF            ┆ null                    ┆ null                         ┆ f      ┆ 1         ┆ 385153371            ┆ 734881840    ┆ root                          │
            │ 8         ┆ 0.0      ┆ Basic cell groups and regions ┆ grey    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1           ┆ [997, 8]          ┆ BFDAE3            ┆ null                    ┆ null                         ┆ f      ┆ 2         ┆ 2244697386           ┆ 734881840    ┆ Basic cell groups and regions │
            │ …         ┆ …        ┆ …                             ┆ …       ┆ …        ┆ …           ┆ …             ┆ …      ┆ …                   ┆ …     ┆ …        ┆ …           ┆ …                 ┆ …                 ┆ …                       ┆ …                            ┆ …      ┆ …         ┆ …                    ┆ …            ┆ …                             │
            │ 65        ┆ 715.0    ┆ parafloccular sulcus          ┆ pfs     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1040.0              ┆ 3     ┆ 1        ┆ 1324        ┆ [997, 1024, … 65] ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1325      ┆ 771629690            ┆ 734881840    ┆ parafloccular sulcus          │
            │ 624       ┆ 926.0    ┆ Interpeduncular fossa         ┆ IPF     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1024.0              ┆ 2     ┆ 1        ┆ 1325        ┆ [997, 1024, 624]  ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1326      ┆ 1476705011           ┆ 734881840    ┆ Interpeduncular fossa         │
            │ 304325711 ┆ -2.0     ┆ retina                        ┆ retina  ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1326        ┆ [997, 304325711]  ┆ 7F2E7E            ┆ null                    ┆ null                         ┆ f      ┆ 1327      ┆ 3295290839           ┆ 734881840    ┆ retina                        │
            └───────────┴──────────┴───────────────────────────────┴─────────┴──────────┴─────────────┴───────────────┴────────┴─────────────────────┴───────┴──────────┴─────────────┴───────────────────┴───────────────────┴─────────────────────────┴──────────────────────────────┴────────┴───────────┴──────────────────────┴──────────────┴───────────────────────────────┘

        """
        if version == '2017':
            filename = 'structure_tree_safe_2017.csv'
        elif version == 'old':
            filename = 'structure_tree_safe.csv'
        else:
            raise ValueError('')

        #
        filepath = CCF_CACHE_DIRECTORY / filename
        if not filepath.exists():
            cls._request_structure_tree(filepath, filename)

        ret = (pl.read_csv(filepath)
               .with_columns(pl.col('parent_structure_id').fill_null(-1))
               .with_columns(pl.col('structure_id_path').map_elements(lambda it: tuple(map(int, it[1:-1].split('/'))))))
        #
        if '2017' in Path(filepath).name:
            ret = ret.with_columns(pl.col('atlas_id').fill_null(-2))
        else:
            ret = ret.with_columns(pl.col('name').alias('safe_name'))

        return ret

    @classmethod
    def _request_structure_tree(cls, dest: PathLike, filename: str) -> Path:
        url = f'https://raw.githubusercontent.com/cortex-lab/allenCCF/master/{filename}'
        resp = download_with_tqdm(url)

        if resp.status_code == 200:
            df = pd.read_csv(io.BytesIO(resp.content))
            pl.from_pandas(df).write_csv(dest)
            fprint(f'DOWNLOAD! {filename} in {dest.parent}', vtype='io')

        else:
            raise RuntimeError('download structure FAIL!')

        return dest

    def structure_tree(self) -> allensdk.core.structure_tree.StructureTree:
        return self.reference.get_structure_tree(structure_graph_id=self.STRUCTURE_GRAPH_ID)

    def get_structures_by_name(self, name: list[str]) -> list[StructureTreeDict]:
        return self.structure_tree().get_structures_by_name(name)

    # ========== #
    # Annotation #
    # ========== #

    @classmethod
    def load_slice_view(cls,
                        source: ALLEN_SOURCE_TYPE,
                        plane_type: PLANE_TYPE,
                        resolution: int = 10) -> AbstractSliceView:
        """
        Load the slice view

        :param source: ``ALLEN_SOURCE_TYPE``
        :param plane_type: ``PLANE_TYPE``
        :param resolution: resolution in um
        :return: :class:`~neuralib.atlas.view.AbstractSliceView()`
        """

        if source in ('npy', 'nrrd'):
            pattern = f'annotation*{resolution}*.{source}'

            try:
                f = uglob(CCF_CACHE_DIRECTORY, pattern)
            except FileNotFoundError:
                f = cls._request_allen_src(source, resolution, CCF_CACHE_DIRECTORY)

            if source == 'npy':
                data = np.load(f)
            elif source == 'nrrd':
                import nrrd
                data = nrrd.read(f)[0]
            else:
                raise ValueError('')

        elif source == 'template':
            f = CCF_CACHE_DIRECTORY / f'template_volume_{resolution}um.npy'
            if not f.exists():
                f = cls._request_allen_src('template', resolution, CCF_CACHE_DIRECTORY)
            data = np.load(f)

        else:
            raise ValueError('')

        return AbstractSliceView(source, plane_type, resolution, data)

    @classmethod
    def _request_allen_src(cls, src_type: ALLEN_SOURCE_TYPE,
                           resolution: int,
                           dest: PathLike) -> Path:
        """

        :param src_type: ALLEN_SOURCE_TYPE = Literal['npy', 'nrrd', 'template']
                * nrrd: source directly from Alleninstitute
                    Seealso: https://download.alleninstitute.org/informatics-archive/current-release/mouse_ccf/annotation/
                * annotation / template: adapted version from cortex-lab
                    Seealso: https://github.com/cortex-lab/allenCCF
        :param resolution:
        :param dest:
        :return:
        """

        if src_type == 'nrrd':
            from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi
            mcapi = MouseConnectivityApi()
            version = MouseConnectivityApi.CCF_VERSION_DEFAULT
            filename = f'annotation_{resolution}.nrrd'
            mcapi.download_annotation_volume(version, resolution, dest / filename)
            out = Path(dest) / filename
            fprint(f'DOWNLOAD! {filename} in {dest}', vtype='io')

        elif src_type in ('annotation', 'template'):

            if src_type == 'annotation':
                url = 'https://figshare.com/ndownloader/files/44925493'
                filename = 'annotation_volume_10um_by_index.npy'
            elif src_type == 'template':
                url = 'https://figshare.com/ndownloader/files/44925496'
                filename = 'template_volume_10um.npy'
            else:
                raise ValueError('')

            out = Path(dest) / filename
            fprint(f'DOWNLOADING... {filename} from {url}', vtype='io')

            resp = download_with_tqdm(url)
            if resp.status_code == 200:
                fprint(f'DOWNLOAD! {filename} in {dest}', vtype='io')

                with BytesIO(resp.content) as f:

                    data = np.load(f, allow_pickle=True)
                np.save(out, data)

            else:
                raise RuntimeError('download allen src FAIL!')
        else:
            raise ValueError('')

        return out

    # ====== #
    # Others #
    # ====== #

    @classmethod
    def load_structure_voxel(cls, to_pandas=False) -> DataFrame:
        """TODO not test yet"""
        file = CCF_CACHE_DIRECTORY / 'voxel_count_and_differences.csv'
        df = (
            pl.read_csv(file)
            .select(pl.col(['acronym', 'name', 'total_voxel_count']))
            .drop_nulls()
            .sort('total_voxel_count', descending=True)
        )

        if to_pandas:
            df = df.to_pandas()

        return df


# ==================== #

def create_allen_structure_dict(verbose=False) -> dict[str, str]:
    """
    Get the acronym/name pairing from structure_tree.csv

    :return: key: acronym; value: full name
    """
    tree = AllenReferenceWrapper.load_structure_tree()
    tree = tree.select('name', 'acronym').sort('name')

    ret = {
        acry: name
        for name, acry in tree.iter_rows()
    }
    if verbose:
        pprint(ret)

    return ret



from typing import Literal

from typing_extensions import TypeAlias

__all__ = ['Area', 'Source', 'Channel', 'MergeLevel',
           'HEMISPHERE_TYPE']

Area: TypeAlias = str
Source: TypeAlias = str
Channel: TypeAlias = str
MergeLevel: TypeAlias = int

#
HEMISPHERE_TYPE = Literal['ipsi', 'contra', 'both']



from __future__ import annotations

import re
from pathlib import Path
from typing import Any, final, Iterable, TypedDict

import numpy as np
import polars as pl
from typing_extensions import Self

from neuralib.plot.figure import plot_figure
from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint
from neuralib.util.utils import deprecated
from .baselog import Baselog, LOG_SUFFIX, StimlogBase
from .baseprot import AbstractStimProtocol
from .session import Session, SessionInfo, get_protocol_sessions
from .stimulus import StimPattern
from .util import unfold_stimuli_condition, try_casting_number

__all__ = ['RiglogData',
           'StimpyProtocol']


@final
class RiglogData(Baselog):
    """class for handle the riglog file for stimpy **bitbucket/github** version
    (mainly tested in the commits derived from master branch)
    """

    def __init__(self,
                 root_path: PathLike,
                 log_suffix: LOG_SUFFIX = '.riglog',
                 diode_offset: bool = True):

        super().__init__(root_path, log_suffix, diode_offset)

        #
        self.__stimlog_cache: StimlogBase | None = None
        self.__prot_cache: StimpyProtocol | None = None

    @classmethod
    def _cache_asarray(cls, filepath: Path) -> np.ndarray:
        output = filepath.with_name(filepath.stem + '_riglog.npy')

        if not output.exists():
            riglog = np.loadtxt(
                filepath,
                delimiter=',',
                comments=['#', 'None'],
                converters={
                    0: lambda it: float(it[1:]),  # it: parameter name: the text of the first column
                    3: lambda it: float(it[:-1])
                }
            )
            np.save(output, riglog)

        return np.load(output)

    @deprecated(reason='caller rare')
    def session_handler(self, session: str | tuple[str, ...]) -> tuple[float, float]:
        """trigger `select_time_range` for specific time range"""
        strial = self.stimlog_data().session_trials()
        if isinstance(session, str):
            t0, t1 = strial[session].time
        elif isinstance(session, tuple):
            t_all = [strial[s].time for s in session]
            t0, t1 = np.min(t_all), np.max(t_all)
            fprint(f'get trange from multiple sessions, {session}: t0:{t0}, t1:{t1}')
        else:
            raise TypeError(f'{type(session)}')

        self.select_time_range(t0, t1)

        return t0, t1

    def select_time_range(self, t0: float, t1: float):
        """overwrite the `instance attr: data` to given time range"""
        t = self.dat[:, 2] / 1000
        mask = np.logical_and(t0 < t, t < t1)
        self.dat = self.dat[mask]

    @property
    def stimlog_file(self) -> Path:
        return self.riglog_file.with_suffix('.stimlog')

    def stimlog_data(self) -> StimlogBase:
        if self.__stimlog_cache is None:
            if self.version == 'stimpy-git':
                from .stimpy_git import StimlogGit
                self.__stimlog_cache = StimlogGit(self, self.stimlog_file, self._diode_offset)
            elif self.version == 'stimpy-bit':
                self.__stimlog_cache = Stimlog(self, self.stimlog_file, self._diode_offset)
            else:
                raise ValueError(f'unknown version: {self.version}')

        return self.__stimlog_cache

    def get_prot_file(self) -> StimpyProtocol:
        if self.__prot_cache is None:
            self.__prot_cache = StimpyProtocol.load(self.stim_prot_file)

        return self.__prot_cache


class StimConfig(TypedDict, total=False):
    codes: dict[str, int]
    """<EVENT_TYPES>:<NUMBER>"""
    commit_hash: str
    """git commit hash"""
    fields: tuple[str, ...]
    """column repr for the logging"""


@final
class Stimlog(StimlogBase):
    """class for handle the stimlog file for stimpy **bitbucket** version
    (mainly tested in the commits derived from master branch)

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        V = number of acquisition sample pulse (Visual parameters)

        M = number of statemachine pulse
    """
    code_version: str
    log_info: dict[int, str] = {}
    log_header: dict[int, list[str]] = {}

    # vstim: start from code 10 in .stimlog
    v_present_time: np.ndarray
    """time in sec (P,)"""
    v_stim: np.ndarray
    """visual stim_index(P,)"""
    v_trial: np.ndarray
    """trial index(P,)"""
    v_photo: np.ndarray
    """photo diode on-off(P,)"""
    v_contrast: np.ndarray
    """stimulus contrast (P,)"""
    v_ori: np.ndarray
    """direction (P,)"""
    v_sf: np.ndarray
    """spatial frequency (P,)"""
    v_phase: np.ndarray
    """phase in time domain (P,)"""
    v_frame_idx: np.ndarray
    """frame index (P,)"""

    # state machine: start from code 20 in .stimlog
    s_on_v: np.ndarray
    """statemachine sync time on v, use for time sync if diode signal not reliable (M,)"""
    s_present_time: np.ndarray
    """(M,)"""
    s_cycle: np.ndarray
    """(M,)"""
    s_new_state: np.ndarray
    """(M,)"""
    s_old_state: np.ndarray
    """(M,)"""
    s_state_elapsed: np.ndarray
    """(M,)"""
    s_trial_type: np.ndarray
    """(M,)"""

    def __init__(self, riglog: RiglogData,
                 file_path: PathLike,
                 diode_offset: bool = True,
                 sequential_offset: bool = True):
        """
        :param riglog: `RiglogData`
        :param file_path: stimlog filepath
        :param diode_offset: if do the diode offset to sync the time to riglog
        :param sequential_offset: do the sequential offset throughout the recording.
        """

        super().__init__(riglog, file_path)
        self.diode_offset = diode_offset

        # diode
        self._do_sequential_diode_offset = sequential_offset
        self._cache_time_offset: float | np.ndarray | None = None

        self._reset()

    def _reset(self):
        v_present_time = []
        v_stim = []
        v_trial = []
        v_photo = []
        v_contrast = []
        v_ori = []
        v_sf = []
        v_phase = []
        v_stim_idx = []
        s_on_v = []
        s_present_time = []
        s_cycle = []
        s_new_state = []
        s_old_state = []
        s_state_elapsed = []
        s_trial_type = []

        with self.stimlog_file.open() as _f:
            for number, line in enumerate(_f):
                line = line.strip()
                if len(line) == 0:
                    continue

                if line.startswith('#'):
                    self._reset_comment_info(line)
                    continue

                part = line.split(',')

                try:
                    code = int(part[0])

                    if code == 10:
                        v_present_time.append(float(part[1]))

                        try:
                            stim_value = int(part[2])
                        except ValueError:  # 'None'
                            stim_value = -1
                        v_stim.append(stim_value)

                        v_trial.append(int(part[3]))
                        v_photo.append(int(part[4]))

                        if len(part) > 5:
                            v_contrast.append(float(part[5]))
                            v_ori.append(int(float(part[6])))
                            v_sf.append(float(part[7]))
                            v_phase.append(float(part[8]))
                            v_stim_idx.append(int(part[9]))
                        else:
                            # empty list item = -1
                            v_contrast.append(-1)
                            v_ori.append(-1)
                            v_sf.append(-1)
                            v_phase.append(-1)
                            v_stim_idx.append(-1)

                    elif code == 20:
                        s_on_v.append(v_present_time[-1])  # sync s to v
                        s_present_time.append(int(part[1]) / 1000)
                        s_cycle.append(int(part[2]))
                        s_new_state.append(int(part[3]))
                        s_old_state.append(int(part[4]))
                        s_state_elapsed.append(int(part[5]) / 1000)
                        s_trial_type.append(int(part[6]))

                    else:
                        raise ValueError(f'unknown code : {code}')

                except BaseException as e:
                    raise RuntimeError(f'line {number}: {line}') from e

        self.v_present_time = np.array(v_present_time)
        self.v_stim = np.array(v_stim)
        self.v_trial = np.array(v_trial)
        self.v_photo = np.array(v_photo)
        self.v_contrast = np.array(v_contrast)
        self.v_ori = np.array(v_ori)
        self.v_sf = np.array(v_sf)
        self.v_phase = np.array(v_phase)
        self.v_frame_idx = np.array(v_stim_idx)

        self.s_on_v = np.array(s_on_v)
        self.s_present_time = np.array(s_present_time)
        self.s_cycle = np.array(s_cycle)
        self.s_new_state = np.array(s_new_state)
        self.s_old_state = np.array(s_old_state)
        self.s_state_elapsed = np.array(s_state_elapsed)
        self.s_trial_type = np.array(s_trial_type)

    def _reset_comment_info(self, line: str):
        if 'CODES' in line:
            heading, content = line.split(': ')
            iter_codes = content.split(',')
            for pair in iter_codes:
                code, num = pair.split('=')
                code = code.strip()
                value = int(num.strip())
                self.log_info[value] = code

        elif 'VLOG HEADER' in line:
            heading, content = line.split(':')
            self.log_header[10] = content.split(', ')

        elif 'STATE HEADER' in line:
            heading, content = line.split(': ')
            self.log_header[20] = content.split(',')

        elif 'Commit' in line:
            heading, content = line.split(': ')
            self.code_version = content.strip()

    @property
    def exp_start_time(self) -> float:
        tstart = self.v_present_time[0]

        if isinstance(self.time_offset, float):
            return tstart + self.time_offset
        elif isinstance(self.time_offset, np.ndarray):
            return tstart + self.time_offset[0]

    @property
    def exp_end_time(self) -> float:
        tend = self.v_present_time[-1]

        if isinstance(self.time_offset, float):
            return tend + self.time_offset
        elif isinstance(self.time_offset, np.ndarray):
            return tend + self.time_offset[-1]

    @property
    def stim_start_time(self) -> float:
        v_start = np.nonzero(self.v_frame_idx == 1)[0][0]
        tstart = self.v_present_time[v_start]

        if isinstance(self.time_offset, float):
            return tstart + self.time_offset
        elif isinstance(self.time_offset, np.ndarray):
            return tstart + self.time_offset[0]

    @property
    def stim_end_time(self) -> float:
        v_end = np.nonzero(np.diff(self.v_frame_idx) < 0)[0][-1] + 1
        tend = self.v_present_time[v_end]

        if isinstance(self.time_offset, float):
            return tend + self.time_offset
        elif isinstance(self.time_offset, np.ndarray):
            return tend + self.time_offset[-1]

    @property
    def stimulus_segment(self) -> np.ndarray:
        """sti_period"""
        v_start = self.v_frame_idx == 1
        t1 = self.v_present_time[v_start]
        t2 = self.v_present_time[np.nonzero(np.diff(self.v_frame_idx) < 0)[0] + 1]

        if isinstance(self.time_offset, float):
            offset = self.time_offset
        elif isinstance(self.time_offset, np.ndarray):
            offset = np.stack([self.time_offset, self.time_offset], axis=1)
        else:
            raise TypeError('')

        t = np.vstack((t1, t2)).T + offset

        return t

    @property
    def time_offset(self) -> float | np.ndarray:
        if self._cache_time_offset is None:
            self._cache_time_offset = diode_time_offset(self.riglog_data,
                                                        self.diode_offset,
                                                        return_sequential=self._do_sequential_diode_offset)
        return self._cache_time_offset

    @property
    def time_offset_statemachine(self) -> np.ndarray:
        """time offset approach using statemachine"""
        return self.s_present_time - self.s_on_v

    def session_trials(self) -> dict[Session, SessionInfo]:
        return {
            prot.name: prot
            for prot in get_protocol_sessions(self)
        }

    def get_stim_pattern(self) -> StimPattern:
        prot = StimpyProtocol.load(self.stimlog_file.with_suffix('.prot'))
        v_start = self.v_frame_idx == 1
        t = self.stimulus_segment
        dire = self.v_ori[v_start]
        sf = self.v_sf[v_start]
        contrast = self.v_contrast[v_start]
        nr = self.v_stim[v_start]  # sti_nr = 0-71, len: 360

        tf = prot.tf[nr]
        dur = prot['dur'][nr]

        return StimPattern(t, dire, sf, tf, contrast, dur)

    def get_time_profile(self):
        raise NotImplementedError('')


# ================= #
# Diode Time Offset #
# ================= #

class DiodeNumberMismatchError(ValueError):

    def __init__(self):
        super().__init__('Diode numbers are not detected reliably')


class DiodeSignalMissingError(RuntimeError):

    def __init__(self):
        super().__init__('no diode signal were found')


def diode_time_offset(rig: RiglogData,
                      diode_offset: bool = True,
                      return_sequential: bool = True,
                      default_offset_value: float = 0.6) -> float | np.ndarray:
    """
    time offset used in the `old stimpy`
    offset time from screen_time .riglog (diode) to .stimlog
    ** normally stimlog time value are smaller than riglog

    :param rig:
    :param diode_offset: whether correct diode signal
    :param return_sequential: return sequential offset, if False, use mean value across diode pulses
    :param default_offset_value: hardware(rig)-dependent offset value

    :return: tuple of offset average and std value
    """

    stimlog = rig.stimlog_data()
    if not isinstance(stimlog, Stimlog):
        raise TypeError('')

    if not diode_offset:
        fprint('no offset', vtype='warning')
        return default_offset_value

    #
    try:
        t = _diode_offset_sequential(rig, debug_plot=False)
    except DiodeNumberMismatchError as e:
        try:
            first_pulse = _check_if_diode_pulse(rig)
            fprint(f'{repr(e)}, use the first pulse diff for alignment', vtype='warning')
            return first_pulse

        except DiodeSignalMissingError as e:
            fprint(f'{repr(e)}, use default value', vtype='warning')
            return default_offset_value

    #
    avg_t = float(np.mean(t))
    std_t = float(np.std(t))
    if not (0 <= avg_t <= 1):
        fprint(f'{avg_t} too large, might not be properly calculated, check...', vtype='warning')

    fprint(f'avg: {avg_t}, std: {std_t}')

    if return_sequential:
        return t
    else:
        return avg_t


def _check_if_diode_pulse(rig: RiglogData) -> float:
    """only count 1st difference"""
    screen_time = rig.screen_event.time
    stimlog = rig.stimlog_data()
    if not isinstance(stimlog, Stimlog):
        raise TypeError('')
    stimlog_vstart = stimlog.v_present_time[(stimlog.v_frame_idx == 1)]

    try:
        # noinspection PyTypeChecker
        return screen_time[0] - stimlog_vstart[0]
    except IndexError as e:
        raise DiodeSignalMissingError() from e


def _diode_offset_sequential(rig: RiglogData, debug_plot: bool = False) -> np.ndarray:
    stimlog = rig.stimlog_data()
    if not isinstance(stimlog, Stimlog):
        raise TypeError('')

    stimlog_vstart = stimlog.v_present_time[(stimlog.v_frame_idx == 1)]
    riglog_vstart = rig.screen_event.time[0::2]

    if len(riglog_vstart) != len(stimlog_vstart):
        raise DiodeNumberMismatchError()
    else:
        if debug_plot:
            _plot_time_alignment_diode(riglog_vstart, stimlog_vstart)

        return riglog_vstart - stimlog_vstart


def _plot_time_alignment_diode(riglog_screen: np.ndarray,
                               stimlog_time: np.ndarray):
    """Plot time alignment (stimlog time value smaller than riglog)"""
    with plot_figure(None) as ax:
        ax.plot(riglog_screen - stimlog_time)
        ax.set(xlabel='Visual stim #', ylabel='Time diff (s)')


# ======== #
# Protocol #
# ======== #

class StimpyProtocol(AbstractStimProtocol):
    """
    class for handle the protocol file for stimpy **bitbucket/github** version
    (mainly tested in the commits derived from master branch)

    `Dimension parameters`:

        N = numbers of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        C = number of Cycle
    """

    @classmethod
    def load(cls, file: PathLike) -> Self:
        file = Path(file)
        with file.open() as f:
            return cls._load(file.name, f)

    @classmethod
    def loads(cls, content: str, name: str = '<string>') -> Self:
        """load string"""
        return cls._load(name, content.split('\n'))

    @classmethod
    def _load(cls, name: str, content: Iterable[str]) -> Self:
        options = {}
        version = 'stimpy-bit'

        state = 0
        for line in content:
            line = line.strip()

            if len(line) == 0 or line.startswith('#'):
                continue

            # change state to 1 if # stimulus conditions
            if state == 0 and line.startswith('n '):
                header = re.split(' +', line)  # extract header
                data = [[] for _ in range(len(header))]
                state = 1

            elif state == 0:
                idx = line.index('=')
                options[line[:idx].strip()] = try_casting_number(line[idx + 1:].strip())

            elif state == 1:
                parts = re.split(' +', line, maxsplit=len(header))
                rows = unfold_stimuli_condition(parts)

                if len(rows) != 1:
                    version = 'stimpy-git'  # determine

                for r in rows:
                    for i, it in enumerate(r):  # for each col
                        data[i].append(it)
            else:
                raise RuntimeError('illegal state')

        assert len(header) == len(data)

        visual_stimuli = {
            field: list(map(str, data[i])) if field == 'evolveParams' else data[i]
            for i, field in enumerate(header)
        }

        # noinspection PyTypeChecker
        return StimpyProtocol(name, options, pl.DataFrame(visual_stimuli), version)

    @property
    def controller(self) -> str:
        """protocol controller name"""
        return self.options['controller']

    @property
    def is_shuffle(self) -> bool:
        return self.options['shuffle'] == 'True'

    @property
    def background(self) -> float:
        return float(self.options.get('background', 0.5))

    @property
    def start_blank_duration(self) -> int:
        return self.options.get('startBlankDuration', 5)

    @property
    def blank_duration(self) -> int:
        return self.options['blankDuration']

    @property
    def trial_blank_duration(self) -> int:
        return 2  # TODO not tested/checked value?

    @property
    def end_blank_duration(self) -> int:
        return self.options.get('endBlankDuration', 5)

    @property
    def trial_duration(self) -> int:
        dur = self['dur']
        return np.sum(dur) + len(dur) * self.blank_duration + self.trial_blank_duration

    @property
    def visual_duration(self) -> int:
        return self.trial_duration * self.n_trials

    @property
    def total_duration(self) -> int:
        return self.start_blank_duration + self.visual_duration + self.end_blank_duration

    @property
    def texture(self) -> str:
        """stimulus texture, {circle, sqr... todo check in stimpy}"""
        return self.options['texture']

    @property
    def mask(self) -> str:
        """todo check in stimpy"""
        return self.options['mask']

    class EvoledParameter(dict[str, Any]):
        """handle the ``evolveParams`` header in the protocol file
        only available in **stimpy-bit** version
        """

        def __init__(self, keys: list[str], data: np.ndarray):
            self.__keys = keys
            self.__data = data  # {'phase':['linear',1]}

        def __len__(self) -> int:
            return len(self.__keys)

        def keys(self):
            return set(self.__keys)

        def __contains__(self, o) -> bool:
            return o in self.__keys

        def __getitem__(self, k: str) -> Any:
            if k not in self.__keys:  # i.e., 'phase'
                raise KeyError()

            # TODO only work for 'linear' case
            ret = []
            for d in self.__data:
                ret.append(d[k][1])

            return np.array(ret)

    def evolve_param_headers(self) -> list[str]:
        """Get parameter header which set in the 'evolveParams, i.g., 'phase'

        :return: header list.

        """
        keys = set()
        data = self['evolveParams']
        for d in data:
            d = eval(d)
            keys.update(d.keys())
        return list(keys)

    @property
    def evolve_params(self) -> EvoledParameter:
        """Get value from 'evolveParams'.

        Examples:

        >>> log: StimpyProtocol

        How many parameters in evolveParams

        >>> len(log.evolve_params)

        List parameter in 'evolveParams'

        >>> log.evolve_params.keys()

        Dose parameter 'phase' in 'evolveParams'?

        >>> 'phase' in log.evolve_params

        Get phase value from 'evolveParams'.

        >>> log.evolve_params['phase']

        :return:
        """
        if self.version == 'stimpy-git':
            raise DeprecationWarning('new stimpy has no evolveParams header')

        data = np.array([eval(it) for it in self['evolveParams']])  # cast back to dict
        return self.EvoledParameter(self.evolve_param_headers(), data)

    @property
    def tf(self) -> np.ndarray:
        if self.version == 'stimpy-bit':
            return self.evolve_params['phase']
        elif self.version == 'stimpy-git':
            return self['tf']
        else:
            raise NotImplementedError('')

    def to_dict(self) -> dict[str, Any]:
        ret = {
            'controller': self.controller,
            'displayType': self.options['displayType'],
            'background': self.options['background'],
            'stimulusType': self.options['stimulusType'],
            'nTrials': self.n_trials,
            'shuffle': self.options['shuffle'],
            'blankDuration': self.blank_duration,
            'startBlankDuration': self.start_blank_duration,
            'endBlankDuration': self.end_blank_duration,
            'visual_stimuli': self.visual_stimuli_dataframe
        }

        return ret



"""
Get basic GPU info table and verbose
---------------

- cuda driver (Windows)
- metal backend support (MacOS)

.. code-block:: python

    from neuralib.util.gpu import setup_clogger
    print_gpu_table()


"""

from __future__ import annotations

import dataclasses
import platform
import subprocess
from typing import TypedDict

import GPUtil
import polars as pl
import torch

from neuralib.util.color_logging import setup_clogger
from neuralib.util.table import rich_data_frame_table

__all__ = [
    'print_gpu_table',
    'check_mps_available',
    'check_nvidia_cuda_available'
]


class GPUInfoWin(TypedDict, total=False):
    id: str
    name: str
    driver_version: str | None
    gpu_load: str
    """percentage of GPU usage"""
    total_memory: float
    """in MB"""
    free_memory: float
    used_memory: float
    temperature: float
    """in celsius"""


def _get_gpu_windows() -> list[GPUInfoWin]:
    ret = []
    gpus = GPUtil.getGPUs()
    for gpu in gpus:
        ret.append(
            GPUInfoWin(
                id=gpu.id,
                name=gpu.name,
                driver_version=gpu.driver,
                gpu_load=gpu.load,
                total_memory=gpu.memoryTotal,
                free_memory=gpu.memoryFree,
                used_memory=gpu.memoryUsed,
                temperature=gpu.temperature
            )
        )
    return ret


class GPUInfoMac(TypedDict, total=False):
    Chipset_Model: str
    Type: str
    Bus: str
    VRAM: str
    Vendor: str
    Device_ID: str
    Revision_ID: str
    Metal_Support: str

    mps_available: bool


def _get_gpu_mac() -> GPUInfoMac:
    """get mac gpu info from subprocess"""
    output = subprocess.check_output(["system_profiler", "SPDisplaysDataType"], universal_newlines=True)
    lines = output.splitlines()

    ret = {}
    cur_gpu = None

    for line in lines:
        line = line.strip()
        if len(line) == 0:
            continue
        elif line.startswith('Graphics/Displays:'):
            cur_gpu = {}
        elif line.startswith('Displays:'):
            break
        elif cur_gpu is not None and line.strip():
            # print(line)
            key, value = line.split(':', 1)
            if value:
                ret[key.strip()] = value.strip()

    ret['mps_available'] = torch.backends.mps.is_available()

    return ret


def print_gpu_table() -> None:
    if platform.system() in ('win32', 'Windows'):
        info = _get_gpu_windows()
        check_nvidia_cuda_available()
    elif platform.system() == 'Darwin':
        info = _get_gpu_mac()
        check_mps_available()
    else:
        raise NotImplementedError('')

    #
    if isinstance(info, dict):
        rich_data_frame_table(info)
    elif isinstance(info, list):
        df = pl.concat([pl.DataFrame(dataclasses.asdict(it)) for it in info])
        rich_data_frame_table(df)
    else:
        raise TypeError('')


def check_mps_available() -> bool:
    logger = setup_clogger()
    if not torch.backends.mps.is_available():
        if not torch.backends.mps.is_built():
            logger.error('MPS not available because pytorch install not built with MPS enable')
        else:
            logger.error('MPS not available because current MacOs version is not 12.3+,'
                         ' or do not have MPS-enabled device on this machine')
        return False
    else:
        logger.info('MPS is available')
        return True


def check_nvidia_cuda_available() -> bool:
    """
    Checks if the GPU driver reacts and otherwise raises a custom error.
    Useful to check before long GPU-dependent processes.
    """
    logger = setup_clogger()
    process = subprocess.Popen('nvidia-smi', shell=True, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, executable="/bin/bash")
    info, error = process.communicate()

    if process.returncode != 0:
        logger.error(f"{error.decode('utf-8')}")
        return False
    else:
        logger.info("nvidia-smi command successful")

    #
    if torch.cuda.is_available():
        logger.info('cuda is available in current env')
    else:
        logger.error('cuda computing in not setup properly')



from __future__ import annotations

import re
from typing import Any, Callable, final

import numpy as np
import polars as pl

from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint
from neuralib.util.utils import deprecated
from .baselog import StimlogBase
from .session import Session, SessionInfo, get_protocol_sessions
from .stimpy_core import RiglogData, StimpyProtocol
from .stimulus import StimPattern

__all__ = ['StimlogGit']


# TODO code 0 options need to be extended
@final
class StimlogGit(StimlogBase):
    """class for handle the stimlog file for stimpy **github** version
    (mainly tested in the commits derived from master branch)

    `Dimension parameters`:

        N = number of visual stimulation (on-off pairs) = (T * S)

        T = number of trials

        S = number of Stim Type

        V = number of acquisition sample pulse (Visual parameters)

        I = number of photo indicator pulse
    """

    log_name: str | None
    code_version: str | None
    code_tags: str | None
    header: list[str]
    rig_trigger: tuple[str, float]
    start_time: float
    end_time: float
    missing_frame: int

    log_info: dict[int, str]
    """{0: <'Gratings', ...>, 1: 'PhotoIndicator', 2: 'StateMachine', 3: 'LogDict'}"""
    log_header: dict[int, list[str]]
    """list of header"""
    log_data: dict[int, list[tuple[Any, ...]]]
    """only for StateMachine(2) and LogDict(3)"""

    # Visual Presentation <Grating, ...>
    v_present_time: np.ndarray  # (V,) float
    v_duration: np.ndarray  # (V,)
    v_contrast: np.ndarray  # (V,)
    v_ori: np.ndarray  # (V,)
    v_phase: np.ndarray  # (V,)
    v_pos: np.ndarray  # shape(V, 2)
    v_size: np.ndarray  # shape(V, 2)
    v_flick: np.ndarray  # (V,)
    v_interpolate: np.ndarray  # (V,) bool
    v_mask: np.ndarray  # (V, ) bool, TODO
    v_sf: np.ndarray  # (V,)
    v_tf: np.ndarray  # (V,)
    v_opto: np.ndarray  # (V,) int
    v_pattern: np.ndarray

    # PhotoIndicator
    p_time: np.ndarray  # (I,)
    p_state: np.ndarray  # (I,) bool
    p_size: np.ndarray  # (I,)
    p_pos: np.ndarray  # (I, 2)
    p_units: np.ndarray  # (I,)  str
    p_mode: np.ndarray  # (I,)  int
    p_frames: np.ndarray  # (I,)  int
    p_enabled: np.ndarray  # (I,)  bool

    def __init__(self, riglog: RiglogData,
                 file_path: PathLike,
                 diode_offset=True):

        super().__init__(riglog, file_path)
        self.diode_offset = diode_offset
        self._cache_time_offset: float | None = None

        self._reset()

    def _reset(self):
        self.log_name = None
        self.code_version = None
        self.code_tags = None
        self.header = []
        self.rig_trigger: tuple[str, float] = ('', 0)
        self.start_time = np.nan  # TODO check, use for sync?
        self.end_time = np.nan  # TODO check
        self.missing_frame = 0
        self.log_info = {}
        self.log_header = {}
        log_data = {}

        with self.stimlog_file.open() as f:
            for line, content in enumerate(f):  # type: int, str
                content = content.strip()
                # `## (CODE VERSION) : (commit hash: 88c4705 - tags: [''])`
                # *: 0-inf; +: 1-inf; ?: 0-1
                m = re.match(r'#+ (.+?)\s*:\s*(.+)', content)
                if m:
                    info_name = m.group(1)
                    info_value = m.group(2)

                    if info_name == 'LOG NAME':
                        self.log_name = info_value
                    elif info_name == 'CODE VERSION':
                        self._reset_code_version(info_value)
                    elif info_name == 'Format':
                        self.header = info_value.split(' ')
                        if self.header != ['source_id', 'time', 'source_infos']:
                            raise RuntimeError('stimlog format changed')
                    elif info_name == 'Rig trigger on':
                        info_value = info_value.split(',')
                        self.rig_trigger = (info_value[0], float(info_value[1]))
                    elif self._reset_log_info(info_name, info_value):
                        pass
                    else:
                        print(f'ignore header line at {line + 1} : {content}')

                elif content.startswith('### START'):
                    self.start_time = float(content[9:].strip())
                elif content.startswith('### END'):
                    self.end_time = float(content[7:].strip())
                elif content.startswith('# Missed') and content.endswith('frames'):
                    self.missing_frame = int(content.split(' ')[2])
                elif content.startswith('#'):
                    print(f'ignore header line at {line + 1} : {content}')

                else:
                    try:
                        self._reset_line(log_data, line + 1, content)
                    except BaseException:
                        raise RuntimeError(f'bad format line at {line + 1} : {content}')

        self.log_data = self._reset_final(log_data)

    def _reset_code_version(self, info_value: str):
        """for ### CODE VERSION,  commit hash and tags"""
        info_value = info_value.strip().split(' ')
        self.code_version = info_value[info_value.index('hash:') + 1]
        self.code_tags = info_value[info_value.index('tags:') + 1]
        if self.code_version == 'None':
            self.code_version = None
        if self.code_tags == 'None':
            self.code_tags = None

    def _reset_log_info(self, code: str, info_value: str):
        try:
            code = int(code)
        except ValueError:
            return False

        name, header = info_value.split(' ', maxsplit=1)
        header = eval(header)
        self.log_info[code] = name
        self.log_header[code] = header
        return True

    def _reset_line(self, log_data: dict[int, list], line: int, content: str):
        message: str
        value: list
        code, time, message = content.split(' ', maxsplit=2)
        code = int(code)
        time = float(time)

        if self.log_info[code] in ('Gratings', 'FunctionBased', 'PhotoIndicator', 'LogDict'):
            value = eval(message)
            if len(self.log_header[code]) != len(value):
                print(f'log category {code} size mismatched at line {line} : {message}')
            else:
                log_data.setdefault(code, []).append((time, *value))

        elif self.log_info[code] == 'StateMachine':
            # [<States.SHOW_STIM: 2>, <States.SHOW_BLANK: 1>]
            value = eval(message.replace('<', '("').replace(':', '",').replace('>', ')'))
            # [("States.SHOW_STIM", 2), ("States.SHOW_BLANK", 1)]
            if len(self.log_header[code]) != len(value):
                print(f'log category {code} size mismatched at line {line} : {message}')
            else:
                log_data.setdefault(code, []).append((time, *value))

        else:
            print(f'unknown log category : at line {line} : {content}')

    def _reset_final(self, log_data: dict[int, list]) -> dict[int, list]:
        remove_code = []
        for code, content in log_data.items():
            if self.log_info[code] in ('Gratings', 'FunctionBased'):
                self.v_present_time = np.array([it[0] for it in content])
                self.v_duration = np.array([it[1] for it in content])
                self.v_contrast = np.array([it[2] for it in content])
                self.v_ori = np.array([it[3] for it in content])
                self.v_phase = np.array([it[4] for it in content])
                self.v_pos = np.array([it[5] for it in content])
                self.v_size = np.array([it[6] for it in content])
                self.v_flick = np.array([it[7] for it in content])
                self.v_interpolate = np.array([it[8] for it in content], dtype=bool)
                self.v_mask = np.array([it[9] is not None for it in content], dtype=bool)
                self.v_sf = np.array([it[10] for it in content])
                self.v_tf = np.array([it[11] for it in content])
                self.v_opto = np.array([it[12] for it in content], dtype=int)
                self.v_pattern = np.array([it[13] for it in content])

                remove_code.append(code)

            elif self.log_info[code] == 'PhotoIndicator':
                self.p_time = np.array([it[0] for it in content])
                self.p_state = np.array([it[1] for it in content], dtype=bool)
                self.p_size = np.array([it[2] for it in content])
                self.p_pos = np.array([it[3] for it in content])
                self.p_units = np.array([it[4] for it in content])
                self.p_mode = np.array([it[5] for it in content], dtype=int)
                self.p_frames = np.array([it[6] for it in content], dtype=int)
                self.p_enabled = np.array([it[7] for it in content], dtype=bool)

                remove_code.append(code)

        for code in remove_code:
            del log_data[code]

        return log_data

    def get_visual_presentation_dataframe(self) -> pl.DataFrame:
        """
        ┌────────────┬──────────┬──────────┬─────┬───┬──────┬─────┬──────┬─────────┐
        │ time       ┆ duration ┆ contrast ┆ ori ┆ … ┆ sf   ┆ tf  ┆ opto ┆ pattern │
        │ ---        ┆ ---      ┆ ---      ┆ --- ┆   ┆ ---  ┆ --- ┆ ---  ┆ ---     │
        │ f64        ┆ i64      ┆ i64      ┆ i64 ┆   ┆ f64  ┆ i64 ┆ i64  ┆ str     │
        ╞════════════╪══════════╪══════════╪═════╪═══╪══════╪═════╪══════╪═════════╡
        │ 18.990026  ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 10  ┆ 0    ┆ square  │
        │ 21.000029  ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 10  ┆ 0    ┆ square  │
        │ …          ┆ …        ┆ …        ┆ …   ┆ … ┆ …    ┆ …   ┆ …    ┆ …       │
        │ 619.054972 ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 50  ┆ 0    ┆ square  │
        │ 619.084972 ┆ 10       ┆ 1        ┆ 0   ┆ … ┆ 0.04 ┆ 50  ┆ 0    ┆ square  │
        └────────────┴──────────┴──────────┴─────┴───┴──────┴─────┴──────┴─────────┘
        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(self.v_present_time).alias('time'),
            pl.Series(self.v_duration).alias('duration'),
            pl.Series(self.v_contrast).alias('contrast'),
            pl.Series(self.v_ori).alias('ori'),
            pl.Series(self.v_phase).alias('phase'),
            pl.Series(self.v_pos).alias('pos'),
            pl.Series(self.v_size).alias('size'),
            pl.Series(self.v_flick).alias('flick'),
            pl.Series(self.v_interpolate).alias('interpolate'),
            pl.Series(self.v_mask).alias('mask'),
            pl.Series(self.v_sf).alias('sf'),
            pl.Series(self.v_tf).alias('tf'),
            pl.Series(self.v_opto).alias('opto'),
            pl.Series(self.v_pattern).alias('pattern')
        )
        return df

    def get_photo_indicator_dataframe(self) -> pl.DataFrame:
        """
        ┌────────────┬───────┬──────┬────────────┬───────┬──────┬────────┬────────┐
        │ time       ┆ state ┆ size ┆ pos        ┆ units ┆ mode ┆ frames ┆ enable │
        │ ---        ┆ ---   ┆ ---  ┆ ---        ┆ ---   ┆ ---  ┆ ---    ┆ ---    │
        │ f64        ┆ bool  ┆ i64  ┆ list[i64]  ┆ str   ┆ i64  ┆ i64    ┆ bool   │
        ╞════════════╪═══════╪══════╪════════════╪═══════╪══════╪════════╪════════╡
        │ 18.990026  ┆ false ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
        │ 21.000029  ┆ true  ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
        │ …          ┆ …     ┆ …    ┆ …          ┆ …     ┆ …    ┆ …      ┆ …      │
        │ 607.094955 ┆ false ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
        │ 609.104958 ┆ true  ┆ 60   ┆ [740, 370] ┆ pix   ┆ 0    ┆ 20     ┆ true   │
        └────────────┴───────┴──────┴────────────┴───────┴──────┴────────┴────────┘
        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(self.p_time).alias('time'),
            pl.Series(self.p_state).alias('state'),
            pl.Series(self.p_size).alias('size'),
            pl.Series(self.p_pos).alias('pos'),
            pl.Series(self.p_units).alias('units'),
            pl.Series(self.p_mode).alias('mode'),
            pl.Series(self.p_frames).alias('frames'),
            pl.Series(self.p_enabled).alias('enable')
        )
        return df

    def get_state_machine_dataframe(self) -> pl.DataFrame:
        """
        ┌────────────┬───────────────────────────┬───────────────────────────┐
        │ time       ┆ state                     ┆ prev_state                │
        │ ---        ┆ ---                       ┆ ---                       │
        │ f64        ┆ object                    ┆ object                    │
        ╞════════════╪═══════════════════════════╪═══════════════════════════╡
        │ 18.990026  ┆ ('States.SHOW_BLANK', 1)  ┆ ('States.STIM_SELECT', 0) │
        │ 20.990029  ┆ ('States.SHOW_STIM', 2)   ┆ ('States.SHOW_BLANK', 1)  │
        │ …          ┆ …                         ┆ …                         │
        │ 609.094958 ┆ ('States.SHOW_STIM', 2)   ┆ ('States.SHOW_BLANK', 1)  │
        │ 619.094972 ┆ ('States.STIM_SELECT', 0) ┆ ('States.SHOW_STIM', 2)   │
        └────────────┴───────────────────────────┴───────────────────────────┘
        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(it[0] for it in self.log_data[2]).alias('time'),
            pl.Series(it[1] for it in self.log_data[2]).alias('state'),
            pl.Series(it[2] for it in self.log_data[2]).alias('prev_state'),
        )
        return df

    def get_log_dict_dataframe(self) -> pl.DataFrame:
        """
        ┌────────────┬──────────┬──────────┬──────────────┬────────────┐
        │ time       ┆ block_nr ┆ trial_nr ┆ condition_nr ┆ trial_type │
        │ ---        ┆ ---      ┆ ---      ┆ ---          ┆ ---        │
        │ f64        ┆ i64      ┆ i64      ┆ i64          ┆ i64        │
        ╞════════════╪══════════╪══════════╪══════════════╪════════════╡
        │ 18.990026  ┆ 0        ┆ 0        ┆ 0            ┆ 1          │
        │ 30.990043  ┆ 1        ┆ 0        ┆ 0            ┆ 1          │
        │ …          ┆ …        ┆ …        ┆ …            ┆ …          │
        │ 595.083939 ┆ 48       ┆ 0        ┆ 0            ┆ 1          │
        │ 607.094955 ┆ 49       ┆ 0        ┆ 0            ┆ 1          │
        └────────────┴──────────┴──────────┴──────────────┴────────────┘
        :return:
        """
        df = pl.DataFrame().with_columns(
            pl.Series(it[0] for it in self.log_data[3]).alias('time'),
            pl.Series(it[1] for it in self.log_data[3]).alias('block_nr'),
            pl.Series(it[2] for it in self.log_data[3]).alias('trial_nr'),
            pl.Series(it[3] for it in self.log_data[3]).alias('condition_nr'),
            pl.Series(it[4] for it in self.log_data[3]).alias('trial_type'),
        )

        return df

    def get_column(self,
                   category: int | str,
                   field: str,
                   dtype=float,
                   mapper: Callable[[Any], Any] | None = None) -> np.ndarray:

        if category in ('Gratings', 'FunctionBased', 'PhotoIndicator'):
            raise RuntimeError(f'preprocessed log : {category}:{field}')

        if isinstance(category, str):
            category = self._get_category_code(category)

        field = self._get_field_index(category, field)
        data = [it[field] for it in self.log_data[category]]

        if mapper is not None:
            # i.e., operator.itemgetter(1) for tuple statemachine ('States.SHOW_STIM', 2)...
            data = list(map(mapper, data))

        return np.array(data, dtype=dtype)

    def _get_category_code(self, category: str) -> int:
        for code, name in self.log_info.items():
            if name == category:
                return code
        raise KeyError('')

    def _get_field_index(self, category: int, field: str) -> int:
        header = self.log_header[category]
        if field == 'time':
            return 0
        return header.index(field) + 1

    @property
    def exp_start_time(self) -> float:
        return self.riglog_data.dat[0, 2] / 1000

    @property
    def exp_end_time(self) -> float:
        return self.riglog_data.dat[-1, 2] / 1000

    @property
    def stim_start_time(self) -> float:
        return self.p_time[1] + self.time_offset

    @property
    def stim_end_time(self) -> float:
        return self.v_present_time[-1] + self.time_offset

    @property
    def stimulus_segment(self) -> np.ndarray:
        # ret = self.riglog_data.screen_event[:, 0].reshape(-1, 2)  # directly use riglog time
        # stimlog bug starting from `diode off`, and ending from `diode True`
        # thus remove first point and add last point manually
        _p_time = np.concatenate((self.p_time[1:], np.array([self.v_present_time[-1]])))
        ret = _p_time.reshape(-1, 2) + self.time_offset
        return ret

    def session_trials(self) -> dict[Session, SessionInfo]:
        return {
            prot.name: prot
            for prot in get_protocol_sessions(self)
        }

    @property
    def time_offset(self) -> float:
        if self._cache_time_offset is None:
            # noinspection PyTypeChecker
            self._cache_time_offset = _time_offset(self.riglog_data, self, self.diode_offset)[0]
        return self._cache_time_offset

    def get_stim_pattern(self) -> StimPattern:
        prot = StimpyProtocol.load(self.stimlog_file.with_suffix('.prot'))
        log_nr = self.get_column('LogDict', 'condition_nr').astype(int)

        t = self.stimulus_segment
        ori = np.array([prot['ori'][n] for n in log_nr])
        sf = np.array([prot['sf'][n] for n in log_nr])
        tf = np.array([prot['tf'][n] for n in log_nr])
        contrast = np.array([prot['c'][n] for n in log_nr])
        dur = np.array([prot['dur'][n] for n in log_nr])

        return StimPattern(t, ori, sf, tf, contrast, dur)

    def get_time_profile(self):
        raise NotImplementedError('')


# ========== #

@deprecated(reason='generalized, to sequential offset method')
def _time_offset(rig: RiglogData,
                 stm: StimlogGit,
                 diode_offset=True) -> tuple[float, float]:
    """
    time offset used in the `new stimpy`
    offset time from screen_time .riglog (diode) to .stimlog
    ** normally stimlog time value are larger than riglog

    :param rig:
    :param stm:
    :param diode_offset: whether correct diode signal
    :return: tuple of offset average and std value
    """
    if not isinstance(stm, StimlogGit):
        raise TypeError('')

    screen_time = rig.screen_event.time

    if diode_offset:
        try:
            # new stimpy might be a negative value (stimlog time value larger than riglog)
            # stimlog bug starting from `diode off`, and ending from `diode True`
            # thus remove first point and add last point manually
            _p_time = np.concatenate((stm.p_time[1:], np.array([stm.v_present_time[-1]])))
            offset_t = screen_time[::2] - _p_time[::2]

        except ValueError as e:
            fprint(f'number of diode pulse and stimulus mismatch from {e}', vtype='error')
            fprint(f'use the first pulse diff for alignment', vtype='error')
            offset_t = screen_time[0] - stm.p_time[1]
    else:
        raise NotImplementedError('')

    offset_t_avg = float(np.mean(offset_t))
    offset_t_std = float(np.std(offset_t))

    fprint(f'time offset between stimlog and riglog: {round(offset_t_avg, 3)}')
    fprint(f'offset_std: {round(offset_t_std, 3)}')

    return offset_t_avg, offset_t_std



from __future__ import annotations

import abc
import math
from typing import Final, Union, Tuple

import attrs
import numpy as np
from matplotlib.axes import Axes
from matplotlib.image import AxesImage
from matplotlib.transforms import CompositeGenericTransform
from typing_extensions import Self

from neuralib.atlas.util import PLANE_TYPE, ALLEN_CCF_10um_BREGMA, ALLEN_SOURCE_TYPE
from neuralib.imglib.factory import ImageProcFactory
from neuralib.plot import plot_figure

__all__ = [
    'AbstractSliceView',
    'SlicePlane'
]


class AbstractSliceView(metaclass=abc.ABCMeta):
    source_type: Final[ALLEN_SOURCE_TYPE]
    plane_type: Final[PLANE_TYPE]
    resolution: Final[int]
    """um/pixel"""
    reference: Final[np.ndarray]
    """(AP, DV, ML)"""

    grid_x: Final[np.ndarray]
    grid_y: Final[np.ndarray]

    reference_verbose: str = ''

    def __new__(cls, source_type: ALLEN_SOURCE_TYPE,
                plane: PLANE_TYPE,
                resolution: int,
                reference: np.ndarray):
        if plane == 'coronal':
            return object.__new__(CoronalSliceView)
        elif plane == 'sagittal':
            return object.__new__(SagittalSliceView)
        elif plane == 'transverse':
            return object.__new__(TransverseSliceView)
        else:
            raise ValueError('')

    def __init__(self, source_type: ALLEN_SOURCE_TYPE,
                 plane: PLANE_TYPE,
                 resolution: int,
                 reference: np.ndarray):
        """

        :param source_type:
        :param plane:
        :param resolution:
        :param reference: (AP, DV, ML)
        """
        self.source_type = source_type
        self.plane_type = plane
        self.resolution = resolution
        self.reference = reference

        self.grid_y, self.grid_x = np.mgrid[0:self.height, 0:self.width]

        self._check_attrs()

    def _check_attrs(self):
        if self.resolution == 10:
            assert self.reference.shape == (1320, 800, 1140)
        elif self.resolution == 25:
            assert self.reference.shape == (528, 320, 456)

    @property
    def bregma(self) -> np.ndarray:
        if self.resolution == 10:
            return ALLEN_CCF_10um_BREGMA
        raise NotImplementedError('')

    @property
    def n_ap(self) -> int:
        """number of slices along AP axis"""
        return self.reference.shape[0]

    @property
    def n_dv(self) -> int:
        """number of slices along DV axis"""
        return self.reference.shape[1]

    @property
    def n_ml(self) -> int:
        """number of slices along ML axis"""
        return self.reference.shape[2]

    @property
    @abc.abstractmethod
    def n_planes(self) -> int:
        """number of planes in a specific plane view"""
        pass

    @property
    @abc.abstractmethod
    def width(self) -> int:
        """width (pixel) in a specific plane view"""
        pass

    @property
    @abc.abstractmethod
    def height(self) -> int:
        """height (pixel) in a specific plane view"""
        pass

    @property
    def width_mm(self) -> float:
        """width (um) in a specific plane view"""
        return self.width * self.resolution / 1000

    @property
    def height_mm(self) -> float:
        """height (um) in a specific plane view"""
        return self.height * self.resolution / 1000

    @property
    @abc.abstractmethod
    def reference_point(self) -> int:
        """reference point in a specific plane view. aka, bregma plane index"""
        pass

    @property
    @abc.abstractmethod
    def project_index(self) -> Tuple[int, int, int]:
        """plane(p), x, y of index order in (AP, DV, ML)

        :return: (p, x, y)
        """
        pass

    def plane_at(self, slice_index: int) -> SlicePlane:
        return SlicePlane(slice_index, int(self.width // 2), int(self.height // 2), 0, 0, self)

    def offset(self, h: int, v: int) -> np.ndarray:
        """

        :param h: horizontal plane diff to the center. right side positive.
        :param v: vertical plane diff to the center. bottom side positive.
        :return: (H, W) array
        """
        x_frame = np.round(np.linspace(-h, h, self.width)).astype(int)
        y_frame = np.round(np.linspace(-v, v, self.height)).astype(int)
        return np.add.outer(y_frame, x_frame)

    def plane(self, offset: Union[int, tuple[int, int, int], np.ndarray]) -> np.ndarray:
        """Get image plane.

        :param offset: Array[int, height, width] or tuple (plane, dh, dv)
        :return:
        """
        if isinstance(offset, int):
            offset = np.full_like((self.height, self.width), offset)
        elif isinstance(offset, tuple):
            offset = offset[0] + self.offset(offset[1], offset[2])
        elif not isinstance(offset, np.ndarray):
            raise TypeError(str(type(offset)))

        offset[offset < 0] = 0
        offset[offset > self.n_planes] = self.n_planes - 1

        return self.reference[self.coor_on(offset, (self.grid_x, self.grid_y))]

    def coor_on(self, plane: np.ndarray,
                o: tuple[np.ndarray, np.ndarray]) -> tuple[np.ndarray, ...]:
        """
        map slice point (x, y) at plane *plane* back to volume point (ap, dv, ml)

        :param plane: plane number of array
        :param o: tuple of (x, y)
        :return: (ap, dv, ml)
        """
        pidx, xidx, yidx = self.project_index
        ret = [0, 0, 0]
        ret[pidx] = plane
        ret[xidx] = o[0]
        ret[yidx] = o[1]

        return tuple(ret)


class CoronalSliceView(AbstractSliceView):
    reference_verbose = 'AP'

    @property
    def n_planes(self) -> int:
        return self.n_ap

    @property
    def width(self) -> int:
        return self.n_ml

    @property
    def height(self) -> int:
        return self.n_dv

    @property
    def reference_point(self) -> int:
        return self.bregma[0]

    @property
    def project_index(self) -> Tuple[int, int, int]:
        return 0, 2, 1


class SagittalSliceView(AbstractSliceView):
    reference_verbose = 'ML'

    @property
    def n_planes(self) -> int:
        return self.n_ml

    @property
    def width(self) -> int:
        return self.n_ap

    @property
    def height(self) -> int:
        return self.n_dv

    @property
    def reference_point(self) -> int:
        return self.bregma[2]

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 2, 0, 1  # p=ML, x=AP, y=DV


class TransverseSliceView(AbstractSliceView):
    reference_verbose = 'DV'

    @property
    def n_planes(self) -> int:
        return self.n_dv

    @property
    def width(self) -> int:
        return self.n_ml

    @property
    def height(self) -> int:
        return self.n_ap

    @property
    def reference_point(self) -> int:
        return self.bregma[1]

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 1, 2, 0


@attrs.define
class SlicePlane:
    """2D Wrapper class of *SliceView for specific plane"""
    slice_index: int  # anchor index
    ax: int  # anchor x
    ay: int  # anchor y
    dw: int  # in um
    dh: int
    view: AbstractSliceView

    @property
    def image(self) -> np.ndarray:
        return self.view.plane(self.plane_offset)

    @property
    def plane_offset(self) -> np.ndarray:
        offset = self.view.offset(self.dw, self.dh)
        return self.slice_index + offset - offset[self.ay, self.ax]

    @property
    def reference_value(self) -> float:
        """relative to reference point"""
        factor = 1000 / self.view.resolution
        return round((self.view.reference_point - self.slice_index) / factor, 2)

    def with_offset(self, dw: int, dh: int) -> Self:
        return attrs.evolve(self, dw=dw, dh=dh)

    def with_angle_offset(self, deg_x: float, deg_y: float) -> Self:
        """

        :param deg_x:
        :param deg_y:
        :return:
        """
        rx = np.deg2rad(deg_x)
        ry = np.deg2rad(deg_y)

        dw = int(-self.view.width * math.tan(rx) / 2)
        dh = int(self.view.height * math.tan(ry) / 2)

        return self.with_offset(dw, dh)

    def plot(self, ax: Axes | None = None,
             to_um: bool = True,
             with_annotation: bool = False,
             cbar: bool = False,
             with_title: bool = False,
             affine_transform: bool = False,
             customized_trans: bool = False,
             **kwargs) -> tuple[AxesImage, AxesImage | None, CompositeGenericTransform]:
        """
        Plot slice view

        :param ax:
        :param to_um: whether plot axis as mm or um
        :param with_annotation: if plot annotation
        :param cbar:
        :param with_title:
        :param affine_transform: whether do the affine transformation
        :param customized_trans: whether use customized transformation method. NOTE the Shear transform matrix [2, 0]
                        value might not be able to apply in other plot (shifting)
        :param kwargs: pass through the image imshow args
        :return:
        """

        if ax is None:
            with plot_figure(None) as ax:
                ret = self._plot(ax, to_um, with_annotation, cbar, with_title, affine_transform,
                                 customized_trans, **kwargs)
        else:
            ret = self._plot(ax, to_um, with_annotation, cbar, with_title, affine_transform, customized_trans, **kwargs)

        return ret

    unit: str = None

    def _get_xy_range(self, to_um=True) -> tuple[float, float, float, float]:
        if to_um:
            x0 = -self.view.width_mm / 2 * 1000
            x1 = self.view.width_mm / 2 * 1000
            y0 = self.view.height_mm * 1000
            y1 = 0
            self.unit = 'um'
        else:
            x0 = -self.view.width_mm / 2
            x1 = self.view.width_mm / 2
            y0 = self.view.height
            y1 = 0
            self.unit = 'mm'

        return x0, x1, y0, y1

    def _customized_affine_transform(self) -> np.ndarray:
        # translation
        Y = np.array([[1, 0, 0], [0, 1, -4000], [0, 0, 1]])

        # shear
        tt = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [-0.03 / self.view.width, 0, 1]
        ])

        # translation
        _Y = np.array([[1, 0, 0], [0, 1, 3500], [0, 0, 1]])

        return Y @ tt @ _Y

    def _plot(self, ax: Axes,
              to_um,
              with_annotation,
              cbar,
              with_title,
              affine_transform,
              customized_trans,
              **kwargs) -> tuple[AxesImage, AxesImage | None, CompositeGenericTransform]:

        x0, x1, y0, y1 = self._get_xy_range(to_um)

        #
        if affine_transform:
            import matplotlib.transforms as mtransforms
            #
            if customized_trans:
                aff = mtransforms.Affine2D(self._customized_affine_transform())
            else:
                aff = mtransforms.Affine2D().skew_deg(-20, 0)

            aff_trans = aff + ax.transData
        else:
            aff_trans = ax.transData

        #
        image = self.image.astype(float)
        image[image <= 10] = np.nan
        im_view = ax.imshow(image, cmap='Greys', extent=(x0, x1, y0, y1), clip_on=False, transform=aff_trans, **kwargs)

        #
        if with_annotation:
            from neuralib.atlas.allen import AllenReferenceWrapper
            ann_img = (AllenReferenceWrapper
                       .load_slice_view('npy', self.view.plane_type, self.view.resolution)
                       .plane(self.plane_offset))
            ann = ImageProcFactory(ann_img).covert_grey_scale().edge_detection(10, 0).image

            ann = ann.astype(float)
            ann[ann <= 10] = np.nan

            im_ann = ax.imshow(ann, cmap='binary', extent=(x0, x1, y0, y1), alpha=0.3, clip_on=False,
                               interpolation='none', vmin=0, vmax=255, transform=aff_trans)

        else:
            im_ann = None

        #
        if cbar:
            ax.figure.colorbar(im_view)
        #
        if with_title:
            ax.set_title(f'{self.view.reference_verbose}: {self.reference_value} mm')

        ax.set(xlabel=self.unit, ylabel=self.unit)

        return im_view, im_ann, aff_trans



from __future__ import annotations

import io
import re
from typing import Any

import polars as pl

from neuralib.stimpy.math_eval import evaluate_string

__all__ = [
    'EmptyProtocolError',
    'UnassignedVariableError',
    'UndefinedVariableError',
    'VariableAlreadyDeclaredError',
    #
    'remove_comments_and_strip',
    'lines_to_variables_dict',
    'eval_assignments',
    'eval_dataframe',
    'generate_extended_dataframe',
]


class ProtocolError(ValueError):
    """Custom Error class for protocol errors"""

    def __init__(self, message: str, line: int | None = None):
        """
        :param message: error message
        :param line: line number
        """
        if line is not None:
            message += f' on protocol line {line}'
        super().__init__(message)
        self.line = line


class VariableAlreadyDeclaredError(ProtocolError):
    def __init__(self, var_name: str, line: int = None):
        super().__init__(f"Variable {var_name} already declared", line=line)


class EmptyProtocolError(ProtocolError):
    def __init__(self):
        super().__init__("empty protocol")


class UndefinedVariableError(ProtocolError):
    def __init__(self, var_name: str, line: int = None):
        super().__init__(f"undefined variable {var_name}", line=line)


class UnassignedVariableError(ProtocolError):
    def __init__(self, var_name: str, line: int = None):
        super().__init__(f"unassigned variable {var_name}", line=line)


# ================= #

def remove_comments_and_strip(lines: list[str]) -> list[str]:
    """Extracts assignments for all lines given,
    removes the eventual comment and strip leading/trailing spacings,
    discards lines without assignment"""
    stripped_lines = []
    for line in lines:
        stripped_line, _, _ = line.partition('#')
        stripped_line = stripped_line.strip()
        if stripped_line != '':
            stripped_lines.append(stripped_line)

    return stripped_lines


def lines_to_variables_dict(lines: list[str]) -> dict[str, str]:
    """Identify assignments (=)
    and create dict {variable_name : string_to_eval}"""
    variables_dict = {}
    variables_loc = {}

    current_variable = ''
    current_string = ''

    assign_pattern = re.compile(r'^([A-Za-z0-9_]+)\s*=\s*(.*)\s*$')

    for line, content in enumerate(lines):

        if (match := assign_pattern.match(content)) is not None:
            current_variable = match.group(1).strip()
            current_string = match.group(2).strip()

            if current_variable in variables_dict:
                raise VariableAlreadyDeclaredError(current_variable, line=line + 1)

            variables_dict[current_variable] = current_string
            variables_loc[current_variable] = line + 1
        else:
            if current_string != '':
                current_string += "\n"

            current_string += content.strip()
            variables_dict[current_variable] = current_string

    check_variables_dict_for_errors(variables_dict, variables_loc)
    return variables_dict


def check_variables_dict_for_errors(variables_dict: dict[str, str], variables_loc: dict[str, int]):
    if len(variables_dict) == 0:
        raise EmptyProtocolError()

    for key, value in variables_dict.items():
        if key == '':
            raise UndefinedVariableError(key, line=variables_loc.get(key, None))
        elif value == '':
            raise UnassignedVariableError(key, line=variables_loc.get(key, None))


def eval_assignments(variables_dict: dict[str, str], *,
                     dataframe_var=('visual_stimuli',)) -> dict[str, Any]:
    """Evaluates the dictionary's values and returns the evaluated dict"""
    evaluated_dict = {}
    for key in variables_dict:
        str_val = variables_dict[key]

        if key in dataframe_var:
            dataframe = eval_dataframe(str_val)
            typed_val = generate_extended_dataframe(dataframe)
        else:
            try:
                typed_val = eval(str_val)
            except (ValueError, SyntaxError, NameError):
                typed_val = str_val

        evaluated_dict[key] = typed_val

    return evaluated_dict


# ==================== #
# DataFrame Processing #
# ==================== #

def eval_dataframe(dataframe_string: str) -> pl.DataFrame:
    """Evaluates the dataframe contained in the string
    Returns True and a polar dataframe object if the string contains
    a dataframe, else returns False and None
    """
    dataframe_string = re.sub(r' +', ';', dataframe_string)
    return pl.read_csv(io.StringIO(dataframe_string), has_header=True, separator=';', use_pyarrow=True)


def generate_extended_dataframe(df: pl.DataFrame) -> pl.DataFrame:
    """Undoes the magic n-i one-liner commands and generates needed extra lines. i=0 on line of declaration.

    Example Raw::

        ┌───────┬─────┬─────┬─────┬───┬────────┬───────┬────────┬─────────┐
        │ n     ┆ dur ┆ xc  ┆ yc  ┆ … ┆ ori    ┆ width ┆ height ┆ pattern │
        │ ---   ┆ --- ┆ --- ┆ --- ┆   ┆ ---    ┆ ---   ┆ ---    ┆ ---     │
        │ str   ┆ i64 ┆ i64 ┆ i64 ┆   ┆ str    ┆ i64   ┆ i64    ┆ str     │
        ╞═══════╪═════╪═════╪═════╪═══╪════════╪═══════╪════════╪═════════╡
        │ 0-11  ┆ 3   ┆ 0   ┆ 0   ┆ … ┆ {i}*30 ┆ 200   ┆ 200    ┆ sqr     │
        │ 12-23 ┆ 3   ┆ 0   ┆ 0   ┆ … ┆ {i}*30 ┆ 200   ┆ 200    ┆ sqr     │
        └───────┴─────┴─────┴─────┴───┴────────┴───────┴────────┴─────────┘

    Example Extended::


        ┌─────┬─────┬─────┬─────┬───┬─────┬───────┬────────┬─────────┐
        │ n   ┆ dur ┆ xc  ┆ yc  ┆ … ┆ ori ┆ width ┆ height ┆ pattern │
        │ --- ┆ --- ┆ --- ┆ --- ┆   ┆ --- ┆ ---   ┆ ---    ┆ ---     │
        │ i64 ┆ i64 ┆ i64 ┆ i64 ┆   ┆ i64 ┆ i64   ┆ i64    ┆ str     │
        ╞═════╪═════╪═════╪═════╪═══╪═════╪═══════╪════════╪═════════╡
        │ 0   ┆ 3   ┆ 0   ┆ 0   ┆ … ┆ 0   ┆ 200   ┆ 200    ┆ sqr     │
        │ 1   ┆ 3   ┆ 0   ┆ 0   ┆ … ┆ 30  ┆ 200   ┆ 200    ┆ sqr     │
        │ 2   ┆ 3   ┆ 0   ┆ 0   ┆ … ┆ 60  ┆ 200   ┆ 200    ┆ sqr     │
        │ …   ┆ …   ┆ …   ┆ …   ┆ … ┆ …   ┆ …     ┆ …      ┆ …       │
        │ 22  ┆ 3   ┆ 0   ┆ 0   ┆ … ┆ 660 ┆ 200   ┆ 200    ┆ sqr     │
        │ 23  ┆ 3   ┆ 0   ┆ 0   ┆ … ┆ 690 ┆ 200   ┆ 200    ┆ sqr     │
        └─────┴─────┴─────┴─────┴───┴─────┴───────┴────────┴─────────┘
    """

    if 'n' not in df.columns or df.schema['n'].is_integer():
        return df

    n = pl.col('n')
    ef = df.filter(n.str.contains('-'))
    vf = df.filter(~n.str.contains('-')).with_columns(n.cast(pl.Int64))

    def _extend_row(row: dict[str, str]) -> dict[str, list[str]]:
        a, _, b = row['n'].partition('-')
        a = int(a)
        b = int(b)
        return {
            k: list(range(a, b + 1)) if k == 'n' else (
                _extend_expr(k, v, a, b) if isinstance(v, str) else [v] * (b - a + 1)
            )
            for k, v in row.items()
        }

    do_not_eval = set()
    eval_as_int = set()

    def _extend_expr(k: str, expr: str, a: int, b: int) -> list[str]:
        try:
            if k not in do_not_eval:
                ret = [evaluate_string(expr.replace('{i}', str(i))) for i in range(a, b + 1)]
                eval_as_int.add(k)
                return ret
        except:
            do_not_eval.add(k)

        return [expr.replace('{i}', str(i)) for i in range(a, b + 1)]

    ef = [
        _extend_row(it)
        for it in ef.iter_rows(named=True)
    ]

    vf = vf.with_columns([
        pl.col(col).cast(pl.Int64)
        for col in eval_as_int
    ])

    return pl.concat([vf, pl.concat(list(map(pl.DataFrame, ef)))]).sort(n)



"""
Default Path for Cache
====================

.. code-block:: python

    CACHE_DIRECTORY = Path.home() / '.cache'
    NEUROLIB_CACHE_DIRECTORY = CACHE_DIRECTORY / 'neuralib'

    # ATLAS
    ATLAS_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'atlas'
    CCF_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ccf_2017'
    IBL_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ibl'

"""
from pathlib import Path

__all__ = [
    'CACHE_DIRECTORY',
    'NEUROLIB_CACHE_DIRECTORY',
    #
    'ATLAS_CACHE_DIRECTORY',
    'CCF_CACHE_DIRECTORY',
    'IBL_CACHE_DIRECTORY'
]

CACHE_DIRECTORY = Path.home() / '.cache'
NEUROLIB_CACHE_DIRECTORY = CACHE_DIRECTORY / 'neuralib'

# ATLAS
ATLAS_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'atlas'
CCF_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ccf_2017'
IBL_CACHE_DIRECTORY = ATLAS_CACHE_DIRECTORY / 'ibl'



"""
Dynamic function generator
==========================

:author:
    Ta-Shun Su

"""
import inspect
import textwrap
from typing import Union, Type, Callable, TypeVar, Any

__all__ = ['create_fn',
           'get_func_default']

T = TypeVar('T')
PARA_TYPE = Union[str, tuple[str, Union[str, Type]], tuple[str, Union[None, str, Type], str]]
PARA_TYPE_LIST = list[PARA_TYPE]
RET_TYPE = Union[None, str, Type]


def _create_fn_para_from_func(func: Callable) -> str:
    para = []

    s = inspect.signature(func)
    p = None
    for arg_name, arg in s.parameters.items():

        if arg.annotation is inspect.Parameter.empty:
            arg_type = ''
        elif isinstance(arg.annotation, str):
            arg_type = f': {arg.annotation}'
        else:
            arg_type = f': {arg.annotation.__name__}'

        if arg.kind != p:
            if p == inspect.Parameter.POSITIONAL_ONLY:
                para.append('/')
                if arg.kind == inspect.Parameter.KEYWORD_ONLY:
                    para.append('*')
            elif p == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                if arg.kind == inspect.Parameter.KEYWORD_ONLY:
                    para.append('*')

        if arg.kind == inspect.Parameter.VAR_POSITIONAL:
            arg_prefix = '*'
        elif arg.kind == inspect.Parameter.VAR_KEYWORD:
            arg_prefix = '**'
        else:
            arg_prefix = ''

        if arg.default is inspect.Parameter.empty:
            arg_default = ''
        else:
            arg_default = f'={repr(arg.default)}'

        para.append(f'{arg_prefix}{arg_name}{arg_type}{arg_default}')

        p = arg.kind

    if s.return_annotation is inspect.Parameter.empty:
        ret_ann = ''
    elif isinstance(s.return_annotation, str):
        ret_ann = f'-> {s.return_annotation}'
    else:
        ret_ann = f'-> {s.return_annotation.__name__}'

    para = ', '.join(para)
    return f'({para}){ret_ann}'


def _create_fn_para_from_list(sign: Union[PARA_TYPE_LIST, tuple[PARA_TYPE_LIST, RET_TYPE]]) -> str:
    para = []

    if isinstance(sign, list):
        ret_ann = ''
    else:
        sign, ret_type = sign
        if ret_type is None:
            ret_ann = '-> None'
        elif isinstance(ret_type, str):
            ret_ann = f'-> {ret_type}'
        elif isinstance(ret_type, type):
            ret_ann = f'-> {ret_type.__name__}'
        else:
            raise TypeError()

    for arg in sign:
        if isinstance(arg, str):
            para.append(arg)
        elif len(arg) == 2:
            arg_name, arg_type = arg
            if isinstance(arg_type, str):
                para.append(f'{arg_name}: {arg_type}')
            elif isinstance(arg_type, type):
                para.append(f'{arg_name}: {arg_type.__name__}')
            else:
                raise TypeError()
        elif len(arg) == 3:
            arg_name, arg_type, default = arg
            if arg_type is None:
                para.append(f'{arg_name}={default}')
            elif isinstance(arg_type, str):
                para.append(f'{arg_name}: {arg_type}={default}')
            elif isinstance(arg_type, type):
                para.append(f'{arg_name}: {arg_type.__name__}={default}')
            else:
                raise TypeError()
        else:
            raise TypeError()

    para = ', '.join(para)
    return f'({para}){ret_ann}'


def create_fn(name: str,
              sign: Union[PARA_TYPE_LIST, tuple[PARA_TYPE_LIST, RET_TYPE], Callable],
              body: str = 'pass', *,
              globals=None,
              locals: dict[str, Any] = None) -> Callable[..., T]:
    """

    Example:

    >>> add = create_fn('add', (['a', 'b'], int), 'return a + b')
    >>> add(1, 2)
    3
    >>> def add_sign(a: int, b:int) -> int:
    ...     pass
    >>> add = create_fn('add', add_sign, 'return a + b')
    >>> add(1, 2)
    3

    Signature Example

    >>> def f(a, b:int, c=0, d:int=1) -> int:
    (['a', ('b', int), ('c', None, '0'), ('d', int, '1')], int)

    reference: dataclasses._create_fn

    :param name:
    :param sign: `([arg_name|(arg_name, arg_type)|(arg_name, arg_type?, str(default)),...], ret_type)`
    :param body:
    :param globals:
    :param locals:
    :return:
    """
    locals = locals or {}

    if inspect.isfunction(sign):
        sign = _create_fn_para_from_func(sign)
    else:
        sign = _create_fn_para_from_list(sign)

    func = textwrap.indent(f'def {name}{sign}:\n' + textwrap.indent(body, prefix='  '), '  ')

    local_params = ', '.join(locals.keys())
    code = f'def __neuralib_dynamic_generate_function__({local_params}):\n{func}\n  return {name}'

    # print(code)
    namespace = {}
    exec(code, globals, namespace)
    return namespace['__neuralib_dynamic_generate_function__'](**locals)


def get_func_default(f: Callable, arg: str) -> Any:
    import inspect
    return inspect.signature(f).parameters[arg].default



from __future__ import annotations

import datetime
from pathlib import Path
from typing import Literal, TypedDict, final

import attrs
import numpy as np
from numpy.typing import NDArray

from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint

__all__ = [
    'SIGNAL_TYPE',
    'CALCIUM_TYPE',
    'Suite2pGUIOptions',
    'Suite2pRoiStat',
    'Suite2PResult'
]

SIGNAL_TYPE = Literal["df_f", "spks"]
CALCIUM_TYPE = Literal['baseline', 'transient']


class Suite2pGUIOptions(TypedDict, total=False):
    """ Suite2p GUI setting.

    .. seealso:: `<https://suite2p.readthedocs.io/en/latest/settings.html>`_"""
    look_one_level_down: float
    fast_disk: str
    delete_bin: bool
    mesoscan: bool
    bruker: bool
    h5py: list
    h5py_key: str
    save_path0: str
    save_folder: str
    subfolders: list
    move_bin: bool
    nplanes: int
    nchannels: int
    functional_chan: int
    tau: float
    fs: float
    force_sktiff: bool
    frames_include: int
    multiplane_parallel: float
    preclassify: float
    save_mat: bool
    save_NWB: float
    combined: float
    aspect: float
    do_bidiphase: bool
    bidiphase: float
    bidi_corrected: bool
    do_registration: int
    two_step_registration: float
    keep_movie_raw: bool
    nimg_init: int
    batch_size: int
    maxregshift: float
    align_by_chan: int
    reg_tif: bool
    reg_tif_chan2: bool
    subpixel: int
    smooth_sigma_time: float
    smooth_sigma: float
    th_badframes: float
    norm_frames: bool
    force_refImg: bool
    pad_fft: bool
    nonrigid: bool
    block_size: tuple[int, int]
    snr_thresh: float
    maxregshiftNR: float
    oneP_reg: bool  # 1Preg
    spatial_hp: int
    spatial_hp_reg: float
    spatial_hp_detect: int
    pre_smooth: float
    spatial_taper: float
    roidetect: bool
    spikedetect: bool
    anatomical_only: float
    sparse_mode: bool
    diameter: float
    spatial_scale: float
    connected: bool
    nbinned: int
    max_iterations: int
    threshold_scaling: float
    max_overlap: float
    high_pass: float
    denoise: bool
    soma_crop: bool
    neuropil_extract: bool
    inner_neuropil_radius: float
    min_neuropil_pixels: int
    lam_percentile: float
    allow_overlap: bool
    use_builtin_classifier: bool
    classifier_path: int
    chan2_thres: float
    baseline: str
    win_baseline: float
    sig_baseline: float
    prctile_baseline: float
    neucoeff: int
    suite2p_version: str
    data_path: list[str]
    sbx_ndeadcols: int
    input_format: str
    save_path: str
    ops_path: str
    reg_file: str
    filelist: list[str]
    nframes_per_folder: np.ndarray
    sbx_ndeadrows: int
    meanImg: np.ndarray
    meanImg_chan2: np.ndarray  # if chan_2
    nframes: int
    Ly: int
    Lx: int
    date_proc: datetime.datetime
    refImg: np.ndarray
    rmin: int
    rmax: int
    yblock: list[np.ndarray]
    xblock: list[np.ndarray]
    nblocks: list[int]
    NRsm: np.ndarray
    yoff: np.ndarray
    xoff: np.ndarray
    corrXY: np.ndarray
    yoff1: np.ndarray
    xoff1: np.ndarray
    corrXY1: np.ndarray
    badframes: np.ndarray
    yrange: list[int]
    xrange: list[int]
    tPC: np.ndarray
    regPC: np.ndarray
    regDX: np.ndarray
    Lyc: int
    Lxc: int
    max_proj: np.ndarray
    Vmax: np.ndarray
    ihop: np.ndarray
    Vsplit: np.ndarray
    Vcorr: np.ndarray
    Vmap: list[np.ndarray]
    spatscale_pix: np.ndarray
    meanImgE: np.ndarray
    timing: dict[str, float]


class Suite2pRoiStat(TypedDict, total=False):
    """Suite2p GUI imaging.

    .. seealso:: `<https://suite2p.readthedocs.io/en/latest/outputs.html#stat-npy-fields>`_
    """
    ypix: np.ndarray
    xpix: np.ndarray
    lam: np.ndarray
    med: list[int, int]
    footprint: float
    mrs: float
    mrs0: float
    compact: float
    solidity: float
    npix: int
    npix_soma: int
    soma_crop: np.ndarray
    overlap: np.ndarray
    radius: float
    aspect_ratio: float
    npix_norm_no_crop: float
    npix_norm: float
    skew: float
    std: float


# ============== #
# Suite2P Result #
# ============== #

@final
@attrs.frozen
class Suite2PResult:
    """suite2p result container

    `Dimension parameters`:

        N: number of neurons

        F: number pf frames
    """
    directory: Path
    """directory contain all the s2p output files"""

    F: np.ndarray
    """transient activity 2D array ``(N, F)``"""

    FNeu: np.ndarray
    """neuropil activity 2D array ``(N, F)``"""

    spks: np.ndarray
    """deconvolved activity 2D array ``(N, F)``"""

    stat: NDArray[Suite2pRoiStat]
    """GUI imaging after registration, i.e., x, ypixel ...: np.ndarray with shape: ``(N, )``"""

    ops: Suite2pGUIOptions
    """GUI options"""

    iscell: np.ndarray
    """cell probability for each ROI ``(N, 2)``"""

    redcell: np.ndarray | None
    """red cell probability 2D array ``(N, 2)``"""

    redcell_threshold: float | None
    """red cell probability threshold"""

    runtime_frate_check: float | None = attrs.field(default=None)
    """If not None, check frame rate lower bound"""

    def __attrs_post_init__(self):
        if self.runtime_frate_check is not None:
            self._check_frame_rate()

    def _check_frame_rate(self):
        """User specific check suite2p config is correct"""
        if self.fs * self.n_plane > self.runtime_frate_check:
            fprint(f'the fr: {self.fs} and n_etl: {self.n_plane} might not set properly in suite2p,'
                   f'check output ops.json', vtype='error')
            raise RuntimeError('fs and n_plane are not set properly in suite2p')

    @classmethod
    def launch_gui(cls, directory: PathLike) -> None:
        """
        launch the suite2p GUI

        :param directory: directory contain all the s2p output files. e.g., <SUITE2P_OUTPUT>/suite2p/plane<P>
        :return:
        """
        from suite2p.gui import gui2p

        if not isinstance(directory, Path):
            directory = Path(directory)
        gui2p.run(str(directory / 'stat.npy'))

    @classmethod
    def load(
            cls,
            directory: PathLike,
            cell_prob: bool | float = 0.5,
            red_cell_threshold: float = 0.65,
            channel: int = 0,
            runconfig_frate: float | None = 30.0,
    ) -> Suite2PResult:
        """
        Load suite2p result from directory

        :param directory: directory contain all the s2p output files.
                e.g., <SUITE2P_OUTPUT>/suite2p/plane<P>
        :param cell_prob: cell probability,
                    bool type: use the binary criteria in GUI output
                    float type: value in ``iscell[:, 1]``
        :param red_cell_threshold: red cell threshold
        :param channel: channel (PMT) number for the functional channel.
                    i.e., 0 if GCaMP, 1 if jRGECO in scanbox setting
        :param runconfig_frate: if not None, check frame rate lower-bound to make sure the s2p runconfig
        :return: :class:`Suite2PResult`
        """
        if not isinstance(directory, Path):
            directory = Path(directory)

        if channel == 0:
            F = np.load(directory / 'F.npy', allow_pickle=True)
            FNeu = np.load(directory / 'Fneu.npy', allow_pickle=True)
            spks = np.load(directory / 'spks.npy')
            stat = np.load(directory / 'stat.npy', allow_pickle=True)
            ops = np.load(directory / 'ops.npy', allow_pickle=True).tolist()
            iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
            r = directory / 'redcell.npy'
            if r.exists():
                redcell = np.load(r, allow_pickle=True)
            else:
                redcell = None

        elif channel == 1:  # second channel
            F = np.load(directory / 'F_chan2.npy', allow_pickle=True)
            FNeu = np.load(directory / 'Fneu_chan2.npy', allow_pickle=True)
            spks = np.load(directory / 'spks.npy')
            stat = np.load(directory / 'stat.npy', allow_pickle=True)
            ops = np.load(directory / 'ops.npy', allow_pickle=True).tolist()
            iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
            redcell = None
        else:
            raise IndexError(f'{channel} unknown')

        #
        if cell_prob is True:
            x = iscell[:, 0] == 1
            F = F[x]
            FNeu = FNeu[x]
            spks = spks[x]
            stat = stat[x]
            iscell = iscell[x]
            if redcell is not None:
                redcell = redcell[x]

        elif isinstance(cell_prob, float):
            x = iscell[:, 1] >= cell_prob
            F = F[x]
            FNeu = FNeu[x]
            spks = spks[x]
            stat = stat[x]
            iscell = iscell[x]
            if redcell is not None:
                redcell = redcell[x]
        else:
            raise TypeError(f'invalid type: {type(cell_prob)}')

        return Suite2PResult(
            directory,
            F,
            FNeu,
            spks,
            stat,
            ops,
            iscell,
            redcell,
            red_cell_threshold if channel == 1 else None,
            runconfig_frate
        )

    @property
    def has_chan2(self) -> bool:
        """if has a second channel"""
        if self.ops['nchannels'] == 2:
            return True
        return False

    @property
    def n_neurons(self) -> int:
        """number of neurons after load.
        could be less than GUI ROI number if use higher cell_prob in
        :meth:`~neuralib.calimg.suite2p.core.Suite2PResult.load()`"""
        return self.F.shape[0]

    @property
    def n_frame(self) -> int:
        """number of frame number"""
        return self.F.shape[1]

    @property
    def cell_prob(self) -> np.ndarray:
        """probability that the ROI is a cell based on the default classifier"""
        return self.iscell[:, 1]

    @property
    def n_red_neuron(self) -> int:
        """number of identified neuron based on red cell threshold"""
        if self.has_chan2:
            return np.count_nonzero(self.red_cell_prob >= self.redcell_threshold)
        else:
            raise RuntimeError('no channel 2')

    @property
    def red_cell_prob(self) -> np.ndarray | None:
        """red cell probability"""
        if self.redcell is None:
            return None
        return self.redcell[:, 1]

    @property
    def signal_baseline(self) -> float:
        """Gaussian filter width in seconds"""
        return self.ops['sig_baseline']

    @property
    def window_baseline(self) -> float:
        """window for max/min filter in seconds"""
        return self.ops['win_baseline']

    @property
    def fs(self) -> float:
        """suite2p approximate frame rate per plane, exact value should be checked in .sbx or .mat"""
        return self.ops['fs']

    @property
    def neucoeff(self) -> float:
        """neuropil coefficient, normally should be ~0.7"""
        return self.ops['neucoeff']

    @property
    def prctile_baseline(self) -> float:
        """percentile of trace to use as baseline if ops['baseline'] = constant_percentile"""
        return self.ops['prctile_baseline']

    @property
    def n_plane(self) -> int:
        """number of optical plane"""
        return self.ops['nplanes']

    @property
    def image_width(self) -> int:
        """image width (in pixel)"""
        return self.ops['Lx']

    @property
    def image_height(self) -> int:
        """image height (in pixel)"""
        return self.ops['Ly']

    @property
    def image_mean(self) -> np.ndarray:
        """mean image for chan0(1st)"""
        return self.ops['meanImg'].T

    @property
    def image_mean_ch2(self) -> np.ndarray:
        """mean image for chan1(2nd)"""
        return self.ops['meanImg_chan2'].T

    @property
    def indicator_tau(self) -> float:
        """The timescale of the sensor (in seconds)"""
        return self.ops['tau']

    @property
    def rigid_x_offsets(self) -> np.ndarray:
        """x-shifts of recording at each timepoint (F, )"""
        return self.ops['xoff']

    @property
    def rigid_y_offsets(self) -> np.ndarray:
        """y-shifts of recording at each timepoint (F, )"""
        return self.ops['yoff']

    @property
    def rigid_xy_offset(self) -> np.ndarray:
        """peak of phase correlation between frame and reference image at each timepoint"""
        return self.ops['corrXY']

    @property
    def nonrigid_x_offsets(self) -> np.ndarray:
        """(frames, block_size)"""
        return self.ops['xoff1']

    @property
    def nonrigid_y_offsets(self) -> np.ndarray:
        return self.ops['yoff1']

    @property
    def nonrigid_xy_offsets(self) -> np.ndarray:
        return self.ops['corrXY1']

    @classmethod
    def load_total_neuron_number(cls,
                                 directory: Path,
                                 cell_prob: float | None = 0.5) -> int:
        """
        load number of neuron based on iscell.npy

        :param directory: directory contains the iscell.npy
        :param cell_prob: cell probability,
                    bool type: use the binary criteria in GUI output
                    float type: value in ``iscell[:, 1]``
        :return:
        """
        iscell = np.load(directory / 'iscell.npy', allow_pickle=True)
        if cell_prob is None:
            return np.count_nonzero(iscell[:, 0] == 1)
        else:
            return np.count_nonzero(iscell[:, 1] >= cell_prob)

    def get_rois_pixels(self) -> np.ndarray:
        """ROIs pixel (N, 2)"""
        ret = np.zeros((self.n_neurons, 2))
        for i in range(self.n_neurons):
            x = np.mean(self.stat[i]['xpix'])
            y = np.mean(self.stat[i]['ypix'])
            ret[i] = np.array([x, y])

        return ret



from __future__ import annotations

from pathlib import Path
from typing import Any

import numpy as np
import sbxreader

from neuralib.argp import int_tuple_type
from neuralib.calimg.scanbox.core import SBXInfo
from neuralib.imglib.labeller import SequenceLabeller
from neuralib.util.util_type import PathLike
from neuralib.util.utils import uglob

__all__ = ['SBXViewer']


class SBXViewer:
    """wrapper for sbxreader

    `Dimension parameters`:

        F = number of frames

        P = number of optical planes

        C = number of PMT channels

        W = FOV width

        H = FOV height

    """

    info: SBXInfo
    """:class:`~neuralib.calimg.scanbox.core.SBXInfo`"""

    sbx_map: sbxreader.sbx_memmap
    """(F, P, C, W, H)"""

    def __init__(self, directory: PathLike):
        """
        :param directory: scanbox file (.sbx). In the same directory should contain the corresponding .mat file
        """

        if not Path(directory).is_dir():
            raise NotADirectoryError(f'{directory}')

        sbx = uglob(directory, '*.sbx')
        self.sbx_map = sbxreader.sbx_memmap(sbx)

        info = uglob(directory, '*.mat')
        self.info = SBXInfo.load(info)

    @property
    def meta(self) -> dict[str, Any]:
        return self.sbx_map.metadata

    @property
    def version(self) -> int:
        return int(self.info.scanbox_version)

    @property
    def height(self) -> int:
        return self.info.sz[0]

    @property
    def width(self) -> int:
        return self.info.sz[1]

    @property
    def n_planes(self) -> int:
        if not len(self.info.otwave) and self.info.volscan == 0:
            return 1
        elif self.info.volscan == 1:
            return len(self.info.otwave)
        raise RuntimeError('')

    @property
    def n_channels(self) -> int:
        """sbx issue, a bit of hard-coded"""
        if self.info.nchan is None:
            if self.info.channels == 2:
                return 1
            elif self.info.channels == 1:
                return 2
            else:
                raise RuntimeError('')
        else:
            return self.info.nchan

    @property
    def n_frames(self) -> int:
        """number of frames/images"""
        return int(self.info.config.frames / self.n_planes)

    def play(self,
             frames: slice | np.ndarray | None,
             plane: int,
             channel: int):
        """
        Play the selected frames using customized CV2 player.

        See :class:`~neuralib.imglib.labeller.SequenceLabeller`

        :param frames: selected frames. If None, play all sequences
        :param plane: number of optical planes
        :param channel: number of PMT channel
        :return:
        """

        if frames is None:
            frames = np.arange(0, self.n_frames)

        data = self.sbx_map[frames, plane, channel, :, :]
        data = np.asarray(data)

        SequenceLabeller.load_sequences(data).main()

    def to_tiff(self,
                frames: slice | np.ndarray | None,
                plane: int,
                channel: int,
                output: PathLike):
        """
        Convert the selected frames to tiff file

        :param frames: selected frames. If None, convert all sequences
        :param plane: number of optical planes
        :param channel: number of PMT channel
        :param output: output filename
        :return:
        """
        import tifffile

        if frames is None:
            frames = np.arange(0, self.n_frames)
        data = self.sbx_map[frames, plane, channel, :, :]
        tifffile.imwrite(output, data)


def main():
    import argparse
    ap = argparse.ArgumentParser(description='view or save the sbx file, If specify the output using -O, save as tiff'
                                             'otherwise, play.')

    ap.add_argument('-D', '--dir', type=Path, required=True, help='directory containing .sbx/.mat scanbox output',
                    dest='directory')
    ap.add_argument('-F', '--frames', metavar='SLICE', type=int_tuple_type, default=None,
                    help='image sequences slice type', dest='frames')
    ap.add_argument('-P', '--plane', type=int, required=True, help='which optic plane', dest='plane')
    ap.add_argument('-C', '--channel', type=int, required=True, help='which pmt channel', dest='channel')
    ap.add_argument('-O', '--output', default=None, help='tiff output, if None, display the sequence', dest='output')

    opt = ap.parse_args()

    viewer = SBXViewer(opt.directory)
    frames = np.arange(*opt.frames).astype(int) if opt.frames is not None else None

    #
    if opt.output is None:
        viewer.play(frames, opt.plane, opt.channel)
    else:
        viewer.to_tiff(frames, opt.plane, opt.channel, opt.output)


if __name__ == '__main__':
    main()



import textwrap

from neuralib.argp import AbstractParser
from neuralib.argp.dispatch import DispatchOption


class Test(AbstractParser, DispatchOption):
    target: str = DispatchOption.argument(
        '--run'
    )

    @classmethod
    def EPILOG(cls):
        return f"""\
Command (--run)
{textwrap.indent(Test.parser_command_epilog(), "  ")}
"""

    def run(self):
        self.invoke_command(self.target)

    @DispatchOption.dispatch('A')
    def run_a(self):
        """doc A."""
        print('a')

    @DispatchOption.dispatch('B')
    def run_b(self):
        """doc B"""
        print('b')


if __name__ == '__main__':
    Test().main()



from __future__ import annotations

import json
from pathlib import Path

import numpy as np

__all__ = ['JsonEncodeHandler',
           #
           'load_from_json',
           'save_to_json']


class JsonEncodeHandler(json.JSONEncoder):
    """extend from the JSONEncoder class and handle the conversions in a default method"""

    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, Path):
            return str(obj)

        return json.JSONEncoder.default(self, obj)


def load_from_json(filepath: str | Path) -> dict:
    with open(filepath, "r") as file:
        return json.load(file)


def save_to_json(filepath: str | Path, my_dict: dict) -> None:
    with open(filepath, "w") as outfile:
        json.dump(my_dict, outfile, sort_keys=True, indent=4)



from __future__ import annotations

__all__ = ['try_casting_number',
           'unfold_stimuli_condition']


def try_casting_number(value: str, do_eval: bool = False) -> float | int | str:
    """
    try casting of string to numerical value

    :param value: str
    :param do_eval: do built in eval()
    :return:
    """
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            if do_eval:
                try:
                    return eval(value)
                except (SyntaxError, NameError):
                    pass

    return value


def unfold_stimuli_condition(parts: list[str]) -> list[list]:
    """
    unfold the numbers of stimuli (each row in prot file) for parsing

    :param parts: list of str foreach row in the visual stimuli dataframe

        e.g., ['1', '3', '0', '0', '1', '0.04', '0', '0', '200', '200', "{'phase':['linear',1]}"]
    """
    ret = []
    if '-' in parts[0]:
        nstim = list(map(int, parts[0].split('-', maxsplit=1)))
        nstim = tuple((nstim[0], nstim[1] + 1))
        ranging = nstim[1] - nstim[0]  # how many n in each row

        for n in range(*nstim):
            parts[0] = f'{n}'
            ext = []
            for i, it in enumerate(parts):
                if '{i}' in it:
                    it = it.replace('{i}', f'{n - 1}%{ranging}')
                if '{t}' in it:
                    raise NotImplementedError('')

                ext.append(try_casting_number(it, do_eval=True))

            ret.append(ext)

    else:
        for i, it in enumerate(parts):
            ret.append(try_casting_number(it, do_eval=True))

        ret = [ret]
    return ret



import unittest

import numpy as np
import polars as pl
from polars.testing import assert_frame_equal

from neuralib.stimpy.protocol_parser import (
    remove_comments_and_strip,
    lines_to_variables_dict,
    EmptyProtocolError,
    UnassignedVariableError,
    UndefinedVariableError,
    VariableAlreadyDeclaredError,
    eval_assignments,
    eval_dataframe,
    generate_extended_dataframe
)


class TestProtocolParser(unittest.TestCase):
    """Test class for parser"""

    def test_remove_comments_and_strip(self):
        lines = [
            " #4",
            "#2",
            "var_1 = 3 # x",
            "# var_2 = 4 # var_3 = 5 # x",
            "var_2 =",
            "  a b c",
            "  d e f",
        ]
        stripped_prot = remove_comments_and_strip(lines)
        expected_result = [
            "var_1 = 3",
            "var_2 =",
            "a b c",
            "d e f"
        ]
        self.assertListEqual(stripped_prot, expected_result)

    def test_lines_to_variables_dict_success(self):
        assignments = [
            "var_1 = 3",
            "var_2 =",
            "a b c",
            "d>=1 e<=2 f"
        ]  # "d e f"
        variables_dict = lines_to_variables_dict(assignments)
        expected_result = {
            "var_1": "3",
            "var_2": "a b c\nd>=1 e<=2 f"
        }
        self.assertDictEqual(variables_dict, expected_result)

    def test_lines_to_variables_dict_empty_raises_error(self):
        with self.assertRaises(EmptyProtocolError):
            lines_to_variables_dict([])

    def test_lines_to_variables_dict_unassigned_raises_error(self):
        with self.assertRaises(UnassignedVariableError):
            lines_to_variables_dict(["a="])

    def test_lines_to_variables_dict_undefined_variable_raises_error(self):
        with self.assertRaises(UndefinedVariableError):
            lines_to_variables_dict(["=a"])

    def test_lines_to_variables_dict_multiple_declaration_raises_error(self):
        with self.assertRaises(VariableAlreadyDeclaredError):
            lines_to_variables_dict(["a=1", "a=2"])

    def test_eval_assignments(self):
        dummy_assignment_dict = {
            "var1": "1",
            "var2": "2.0",
            "var3": "True",
            "var4": "astring",
            "var5": "[1,2]",
            "var6": "{'v1': 1}",
            "var7": "dur ori other\n1 2 3",
        }
        evaluated_dict = eval_assignments(dummy_assignment_dict, dataframe_var=['var7'])
        self.assertEqual(evaluated_dict["var1"], int(1))
        self.assertEqual(evaluated_dict["var2"], float(2))
        self.assertEqual(evaluated_dict["var3"], True)
        self.assertEqual(evaluated_dict["var4"], "astring")
        self.assertEqual(evaluated_dict["var5"], [1, 2])
        self.assertEqual(evaluated_dict["var6"], {"v1": 1})
        df = pl.DataFrame(data=dict(dur=[1], ori=[2], other=[3]))
        assert_frame_equal(evaluated_dict["var7"], df, check_dtype=False)

    def test_parse_dataframe(self):
        data_frame_line = """\
n   img   g   dur len   xc  yc   c   width   height   mask
0    0    0   2   3    -15  0   1.0   15      15    circle
1    1    0   2   3     15  0   0.5   15      15    circle
2    0    1   2   3     15  0   1.0   15      15    circle
3-4    1    1   2   3    -15  0   0.5   15      15    circle
"""
        df = eval_dataframe(data_frame_line)
        assert_frame_equal(df, pl.DataFrame(dict(
            n=['0', '1', '2', '3-4'],
            img=[0, 1, 0, 1],
            g=[0, 0, 1, 1],
            dur=[2, 2, 2, 2],
            len=[3, 3, 3, 3],
            xc=[-15, 15, 15, -15],
            yc=[0, 0, 0, 0],
            c=[1.0, 0.5, 1.0, 0.5],
            width=[15, 15, 15, 15],
            height=[15, 15, 15, 15],
            mask=['circle', 'circle', 'circle', 'circle'],
        )))

    def test_generate_extended_dataframe(self):
        df = pl.DataFrame(dict(
            n=['0', '2-4', '1', ],
            v=['1', '{i}+3', '2', ],
            u=[0, 1, 2]
        ))
        df = generate_extended_dataframe(df)

        result = pl.DataFrame({
            'n': [0, 1, 2, 3, 4],
            'v': [1, 2, 5, 6, 7],
            'u': [0, 2, 1, 1, 1]
        })

        print(result)
        assert_frame_equal(df, result)


class TestExtendedProtocol(unittest.TestCase):

    def test_variable_i(self):
        dataframe_string = """\
n      dur     xc   yc   c    sf    tf   ori      width  height pattern
0-11  3       0    0    1    0.04  1    {i}*30   200    200    sqr
12-23  3       0    0    1    0.08  1    {i}*30   200    200    sqr
"""

        df = eval_dataframe(dataframe_string)
        print(df)
        df = generate_extended_dataframe(df)
        print(df)

        n_rows = df.shape[0]
        expected_result = pl.DataFrame(
            {
                'n': pl.Series(np.arange(n_rows)),
                'dur': pl.Series(np.full(n_rows, 3)),
                'xc': pl.Series(np.full(n_rows, 0)),
                'yc': pl.Series(np.full(n_rows, 0)),
                'c': pl.Series(np.full(n_rows, 1)),
                'sf': pl.Series([0.04] * 12 + [0.08] * 12),
                'tf': pl.Series(np.full(n_rows, 1)),
                'ori': pl.Series(np.arange(0, n_rows * 30, 30)),
                'width': pl.Series(np.full(n_rows, 200)),
                'height': pl.Series(np.full(n_rows, 200)),
                'pattern': pl.Series(np.full(n_rows, 'sqr')),
            }
        )

        assert_frame_equal(df, expected_result)



import functools
import os
import socket
from datetime import datetime
from pathlib import Path

from neuralib.util.util_verbose import fprint

__all__ = ['slack_bot',
           'send_slack_bot']


def slack_bot(timestamp: bool = True):
    """
    sending error message to slack if any error occurs
    """

    def _decorator(f):
        @functools.wraps(f)
        def _bot(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except BaseException as e:
                text = f'`{socket.gethostname()}` ERROR: {e} in `{Path(f.__code__.co_filename).name}`'

                # find opt as first arg, get detailed information
                if len(args) > 0:
                    opt = args[0]
                    exp_date = getattr(opt, 'EXP_DATE', None)
                    animal = getattr(opt, 'ANIMAL', None)

                    if exp_date is not None and animal is not None:
                        text += f'while running DATA: `{exp_date}_{animal}`'

                    if timestamp:
                        text += f'at `{datetime.today().strftime("%y-%m-%d %H:%M:%S")}`'

                send_slack_bot(text)

                raise e

        return _bot

    return _decorator


def send_slack_bot(text: str):
    env_path = Path(__file__).parent / '.env'
    try:
        import slack
        from dotenv import load_dotenv
        load_dotenv(dotenv_path=env_path)
        try:
            client = slack.WebClient(token=os.environ['SLACK_TOKEN'])
        except KeyError:
            fprint('To get the slack notification, generate token in local machine first', vtype='warning')
        else:
            client.chat_postMessage(channel='#rscvp', text=text)
    except BaseException as e:  # not care, internet disconnected
        fprint(e, vtype='error')



from __future__ import annotations

import abc
import re
from pathlib import Path
from typing import Any, get_args, Final, Literal, final

import numpy as np
import polars as pl
from scipy.interpolate import interp1d
from typing_extensions import Self

from neuralib.util.util_verbose import fprint
from neuralib.util.utils import uglob
from .baselog import CAMERA_TYPE, Baselog
from .stimpy_core import RiglogData
from .stimpy_pyv import PyVlog

__all__ = [
    'CAMERA_VERSION',
    #
    'AbstractCamlog',
    'LabCamlog',
    'PyCamlog'
]

CAMERA_VERSION = Literal['labcam', 'pycam']


class AbstractCamlog(metaclass=abc.ABCMeta):
    """ABC for the camera logging information"""

    root: Path
    """tif root path"""
    frame_id: np.ndarray
    """frame index"""
    timestamp: np.ndarray
    """time stamp of each frame"""
    comment_info: dict[str, Any] = {}
    """# labcams version: 0.2"""
    time_info: list[str] | None = []
    """i.e., # [21-03-02 15:23:08]. Note that time could be duplicated"""

    def __init__(self,
                 root,
                 frame_id,
                 timestamp,
                 comment_info,
                 time_info=None):
        self.root: Final[Path] = root
        self.frame_id: Final[np.ndarray] = frame_id
        self.timestamp: Final[np.ndarray] = timestamp
        self.comment_info: Final[dict[str, Any]] = comment_info
        self.time_info: Final[list[str] | None] = time_info

    def __repr__(self):
        ret = pl.DataFrame().with_columns(
            pl.Series(self.frame_id).alias('frame_id'),
            pl.Series(self.timestamp).alias('time')
        )
        return str(ret)

    @classmethod
    @abc.abstractmethod
    def load(cls, root: Path | str,
             suffix: str = '.camlog') -> Self:
        """
        load/parse the camera log file

        :param root: directory with camera log file
        :param suffix: file suffix
        :return: :class:`AbstractCamlog`
        """
        pass

    @classmethod
    @abc.abstractmethod
    def _parse_comment(cls, content: str) -> None:
        """update the comment_info attr"""
        pass

    @abc.abstractmethod
    def get_camera_time(self,
                        riglog: RiglogData | PyVlog,
                        cam_name: str = '1P_cam') -> np.ndarray:
        pass

    # noinspection PyTypeChecker
    @property
    def nframes(self) -> int:
        return self.frame_id[-1]


# ======= #
# LabCams #
# ======= #

@final
class LabCamlog(AbstractCamlog):

    @classmethod
    def load(cls, root: Path | str,
             suffix: str = '.camlog') -> Self:

        root = Path(root)
        if root.is_dir():
            log_file = uglob(root, f'*{suffix}')
            root = root
        else:
            log_file = root
            root = root.parent

        log_data = []

        with log_file.open() as f:
            for line, content in enumerate(f):
                content = content.strip()
                cls._parse_comment(content)
                if not content.startswith('#'):
                    log_data.append(list(map(float, content.split(','))))

            data = np.array(log_data)

            frame_id = data[:, 0].astype(int)
            timestamp = data[:, 1]

        return LabCamlog(root, frame_id, timestamp, cls.comment_info, cls.time_info)

    @classmethod
    def _parse_comment(cls, content: str) -> None:
        header_pattern = r'#+ (?!.*?\[)#?([^:\n]+):\s*(.+)'
        m = re.match(header_pattern, content)
        if m:
            name, info = m.group(1), m.group(2)
            cls.comment_info.update({f'{name}': info})

        event_pattern = r'# \[(.*?)\]\s*-\s*(.*)'
        if re.match(event_pattern, content):
            cls.time_info.append(content)

    def get_camera_time(self, log: Baselog,
                        cam_name: CAMERA_TYPE = '1P_cam') -> np.ndarray:
        """Interpolate cameralog frames to those recorded by pyvstim.

        :param log: :class:`~neuralib.stimpy.baselog.Baselog`
        :param cam_name: camera name
        :return: 1D camera time array in sec
        """
        if cam_name not in get_args(CAMERA_TYPE):
            raise ValueError(f'{cam_name} unknown')

        cam_pulses = len(log.camera_event[cam_name])

        if cam_pulses != self.frame_id[-1]:
            fprint(f'Loss frame between riglog[{cam_pulses}] vs camlog[{self.frame_id[-1]}]',
                   vtype='warning')

        return interp1d(
            log.camera_event[cam_name].value,
            log.camera_event[cam_name].time,
            fill_value='extrapolate'
        )(self.frame_id)


# ====== #
# PyCams #
# ====== #

@final
class PyCamlog(AbstractCamlog):
    """Pycams log, 2023 AP dev
    TODO not test yet
    """

    @classmethod
    def load(cls, root: Path | str, suffix: str = '.log') -> Self:
        root = Path(root)
        if root.is_dir():
            log_file = uglob(root, f'*{suffix}')
            root = root
        else:
            log_file = root
            root = root.parent

        log_data = []

        with log_file.open() as f:
            for line, content in enumerate(f):
                content = content.strip()
                cls._parse_comment(content)
                if not content.startswith('#'):
                    log_data.append(list(map(float, content.split(','))))

            data = np.array(log_data)

            frame_id = data[:, 0].astype(int)
            timestamp = data[:, 1]

        return PyCamlog(root, frame_id, timestamp, cls.comment_info)

    @classmethod
    def _parse_comment(cls, content: str) -> None:
        match = re.match(r'#+ (.+?)\s*:\s*(.+)', content)
        if match:
            name, value = match.group(1), match.group(2)
            cls.comment_info.update({name: value})

    def get_camera_time(self, riglog: RiglogData, cam_name: CAMERA_TYPE = '1P_cam'):
        raise NotImplementedError('')



from neuralib.argp import parse_command_args
from neuralib.atlas.brainrender import BrainReconstructor, RoisReconstructor, ProbeReconstructor


def main():
    parsers = dict(
        area=BrainReconstructor,
        roi=RoisReconstructor,
        probe=ProbeReconstructor
    )

    parse_command_args(
        parsers=parsers,
        description='brainrender view',
        usage="""
        Example: 
        >> brender area -R VISal,VISam,VISl,VISli,VISp,VISpl,VISpm,VISpor
        >> brender roi -F <FILE>
        >> brender probe -F <FILE> --depth <DEPTH in um>
        """
    )


if __name__ == '__main__':
    main()



from __future__ import annotations

import polars as pl
from bg_atlasapi import BrainGlobeAtlas
from plotly import express as px

from neuralib.util.util_type import PathLike


def plot_sunburst_acronym(source: str = 'allen_mouse_10um',
                          check_latest: bool = True,
                          output: PathLike | None = None):
    """
    plot allen brain structure tree interactive plot

    :param source: allen source name
    :param check_latest: if check the brainglobe api latest version
    :param output: figure output path, otherwise, render interactively
    :return:
    """
    file = BrainGlobeAtlas(source, check_latest=check_latest).root_dir / 'structures.csv'
    df = pl.read_csv(file)

    name = df.select(pl.col('acronym').alias('name'), pl.col('id'), pl.col('parent_structure_id').cast(int))
    xx = name.join(name, left_on='parent_structure_id', right_on='id')
    yy = xx.select(pl.col('name'), pl.col('name_right').alias('parent'))

    data = dict(
        character=yy.get_column('name'),
        parent=yy.get_column('parent'),
    )

    #
    fig = px.sunburst(
        data,
        names='character',
        parents='parent',

    )
    if output is not None:
        fig.write_image(output)
    else:
        fig.show()


if __name__ == '__main__':
    plot_sunburst_acronym()
